import math
from logging import getLogger
from typing import List, NamedTuple, Optional

from eth_abi.packed import encode_abi_packed
from hexbytes import HexBytes
from web3 import Web3
from web3.types import TxParams, Wei

from gnosis.eth.constants import GAS_CALL_DATA_BYTE, NULL_ADDRESS
from gnosis.eth.contracts import (get_proxy_factory_contract,
                                  get_safe_contract, get_safe_V1_0_0_contract)
from gnosis.eth.utils import generate_address_2

logger = getLogger(__name__)


class InvalidERC20Token(Exception):
    pass


class SafeCreate2Tx(NamedTuple):
    salt_nonce: int
    owners: List[str]
    threshold: int
    fallback_handler: str
    master_copy_address: str
    proxy_factory_address: str
    payment_receiver: str
    payment_token: str
    payment: int
    gas: int
    gas_price: int
    payment_token_eth_value: float
    fixed_creation_cost: Optional[int]
    safe_address: str
    safe_setup_data: bytes

    @property
    def payment_ether(self):
        return self.gas * self.gas_price


class SafeCreate2TxBuilder:
    def __init__(self, w3: Web3, master_copy_address: str, proxy_factory_address: str):
        """
        Init builder for safe creation using create2
        :param w3: Web3 instance
        :param master_copy_address: `Gnosis Safe` master copy address
        :param proxy_factory_address: `Gnosis Proxy Factory` address
        """
        assert Web3.isChecksumAddress(master_copy_address)
        assert Web3.isChecksumAddress(proxy_factory_address)

        self.w3 = w3
        self.master_copy_address = master_copy_address
        self.proxy_factory_address = proxy_factory_address
        self.safe_version = get_safe_contract(w3, master_copy_address).functions.VERSION().call()
        if self.safe_version == '1.1.1':
            self.master_copy_contract = get_safe_contract(w3, master_copy_address)
        elif self.safe_version == '1.0.0':
            self.master_copy_contract = get_safe_V1_0_0_contract(w3, master_copy_address)
        else:
            raise ValueError('Safe version must be 1.1.1 or 1.0.0')
        self.proxy_factory_contract = get_proxy_factory_contract(w3, proxy_factory_address)

    def build(self, owners: List[str], threshold: int, salt_nonce: int,
              gas_price: int, fallback_handler: Optional[str] = None, payment_receiver: Optional[str] = None,
              payment_token: Optional[str] = None,
              payment_token_eth_value: float = 1.0, fixed_creation_cost: Optional[int] = None):
        """
        Prepare Safe creation
        :param owners: Owners of the Safe
        :param threshold: Minimum number of users required to operate the Safe
        :param fallback_handler: Handler for fallback calls to the Safe
        :param salt_nonce: Web3 instance
        :param gas_price: Gas Price
        :param payment_receiver: Address to refund when the Safe is created. Address(0) if no need to refund
        :param payment_token: Payment token instead of paying the funder with ether. If None Ether will be used
        :param payment_token_eth_value: Value of payment token per 1 Ether
        :param fixed_creation_cost: Fixed creation cost of Safe (Wei)
        """

        assert 0 < threshold <= len(owners)
        fallback_handler = fallback_handler or NULL_ADDRESS
        payment_receiver = payment_receiver or NULL_ADDRESS
        payment_token = payment_token or NULL_ADDRESS
        assert Web3.isChecksumAddress(payment_receiver)
        assert Web3.isChecksumAddress(payment_token)

        # Get bytes for `setup(address[] calldata _owners, uint256 _threshold, address to, bytes calldata data,
        # address paymentToken, uint256 payment, address payable paymentReceiver)`
        # This initializer will be passed to the ProxyFactory to be called right after proxy is deployed
        # We use `payment=0` as safe has no ether yet and estimation will fail
        safe_setup_data: bytes = self._get_initial_setup_safe_data(owners, threshold, fallback_handler=fallback_handler,
                                                                   payment_token=payment_token,
                                                                   payment_receiver=payment_receiver)

        magic_gas: int = self._calculate_gas(owners, safe_setup_data, payment_token)
        estimated_gas: int = self._estimate_gas(safe_setup_data,
                                                salt_nonce, payment_token, payment_receiver)
        logger.debug('Magic gas %d - Estimated gas %d' % (magic_gas, estimated_gas))
        gas = max(magic_gas, estimated_gas)

        # Payment will be safe deploy cost
        payment = self._calculate_refund_payment(gas,
                                                 gas_price,
                                                 fixed_creation_cost,
                                                 payment_token_eth_value)

        # Now we have a estimate for `payment` so we get initialization data again
        final_safe_setup_data: bytes = self._get_initial_setup_safe_data(owners, threshold,
                                                                         fallback_handler=fallback_handler,
                                                                         payment_token=payment_token, payment=payment,
                                                                         payment_receiver=payment_receiver)

        safe_address = self.calculate_create2_address(final_safe_setup_data, salt_nonce)
        assert int(safe_address, 16), 'Calculated Safe address cannot be the NULL ADDRESS'

        return SafeCreate2Tx(salt_nonce, owners, threshold, fallback_handler,
                             self.master_copy_address, self.proxy_factory_address,
                             payment_receiver, payment_token, payment, gas, gas_price, payment_token_eth_value,
                             fixed_creation_cost, safe_address, final_safe_setup_data)

    @staticmethod
    def _calculate_gas(owners: List[str], safe_setup_data: bytes, payment_token: str) -> int:
        """
        Calculate gas manually, based on tests of previosly deployed safes
        :param owners: Safe owners
        :param safe_setup_data: Data for proxy setup
        :param payment_token: If payment token, we will need more gas to transfer and maybe storage if first time
        :return: total gas needed for deployment
        """
        base_gas = 205000  # Transaction base gas

        # If we already have the token, we don't have to pay for storage, so it will be just 5K instead of 20K.
        # The other 1K is for overhead of making the call
        if payment_token != NULL_ADDRESS:
            payment_token_gas = 55000
        else:
            payment_token_gas = 0

        data_gas = GAS_CALL_DATA_BYTE * len(safe_setup_data)  # Data gas
        gas_per_owner = 20000  # Magic number calculated by testing and averaging owners
        return base_gas + data_gas + payment_token_gas + len(owners) * gas_per_owner

    @staticmethod
    def _calculate_refund_payment(gas: int, gas_price: int, fixed_creation_cost: Optional[int],
                                  payment_token_eth_value: float) -> int:
        if fixed_creation_cost is None:
            # Payment will be safe deploy cost + transfer fees for sending ether to the deployer
            base_payment: int = gas * gas_price
            # Calculate payment for tokens using the conversion (if used)
            return math.ceil(base_payment / payment_token_eth_value)
        else:
            return fixed_creation_cost

    def calculate_create2_address(self, safe_setup_data: bytes, salt_nonce: int):
        proxy_creation_code = self.proxy_factory_contract.functions.proxyCreationCode().call()
        salt = self.w3.keccak(encode_abi_packed(['bytes', 'uint256'], [self.w3.keccak(safe_setup_data), salt_nonce]))
        deployment_data = encode_abi_packed(['bytes', 'uint256'], [proxy_creation_code,
                                                                   int(self.master_copy_address, 16)])
        return generate_address_2(self.proxy_factory_contract.address, salt, deployment_data)

    def _estimate_gas(self, initializer: bytes, salt_nonce: int,
                      payment_token: str, payment_receiver: str) -> int:
        """
        Gas estimation done using web3 and calling the node
        Payment cannot be estimated, as no ether is in the address. So we add some gas later.
        :param initializer: Data initializer to send to GnosisSafe setup method
        :param salt_nonce: Nonce that will be used to generate the salt to calculate
        the address of the new proxy contract.
        :return: Total gas estimation
        """

        # Estimate the contract deployment. We cannot estimate the refunding, as the safe address has not any fund
        gas: int = self.proxy_factory_contract.functions.createProxyWithNonce(self.master_copy_address,
                                                                              initializer, salt_nonce).estimateGas()

        # It's not very relevant if is 1 or 9999
        payment: int = 1

        # We estimate the refund as a new tx
        if payment_token == NULL_ADDRESS:
            # Same cost to send 1 ether than 1000
            gas += self.w3.eth.estimate_gas({'to': payment_receiver, 'value': Wei(payment)})
        else:
            # Top should be around 52000 when storage is needed (funder no previous owner of token),
            # we use value 1 as we are simulating an internal call, and in that calls you don't pay for the data.
            # If it was a new tx sending 5000 tokens would be more expensive than sending 1 because of data costs
            gas += 55000
            # try:
            #     gas += get_erc20_contract(self.w3,
            #                               payment_token).functions.transfer(payment_receiver,
            #                                                                 payment).estimateGas({'from':
            #                                                                                      payment_token})
            # except ValueError as exc:
            #     raise InvalidERC20Token from exc

        return gas

    def _get_initial_setup_safe_data(self, owners: List[str], threshold: int,
                                     fallback_handler: str = NULL_ADDRESS,
                                     payment_token: str = NULL_ADDRESS,
                                     payment: int = 0,
                                     payment_receiver: str = NULL_ADDRESS) -> bytes:
        empty_params: TxParams = {
            'gas': Wei(1),
            'gasPrice': Wei(1),
        }

        if self.safe_version == '1.1.1':
            return HexBytes(self.master_copy_contract.functions.setup(
                owners,
                threshold,
                NULL_ADDRESS,  # Contract address for optional delegate call
                b'',            # Data payload for optional delegate call
                fallback_handler,  # Handler for fallback calls to this contract
                payment_token,
                payment,
                payment_receiver
            ).buildTransaction(empty_params)['data'])
        elif self.safe_version == '1.0.0':
            return HexBytes(self.master_copy_contract.functions.setup(
                owners,
                threshold,
                NULL_ADDRESS,  # Contract address for optional delegate call
                b'',  # Data payload for optional delegate call
                payment_token,
                payment,
                payment_receiver
            ).buildTransaction(empty_params)['data'])
        else:
            raise ValueError('Safe version must be 1.1.1 or 1.0.0')



from django.db import models

from ..models import EthereumAddressField, Sha3HashField, Uint256Field


class EthereumAddress(models.Model):
    value = EthereumAddressField(null=True)


class Uint256(models.Model):
    value = Uint256Field(null=True)


class Sha3Hash(models.Model):
    value = Sha3HashField(null=True)



class SafeServiceException(Exception):
    pass


class CannotEstimateGas(SafeServiceException):
    pass


class CannotRetrieveSafeInfoException(SafeServiceException):
    pass


class InvalidChecksumAddress(SafeServiceException):
    pass


class InvalidPaymentToken(SafeServiceException):
    pass


class InvalidMultisigTx(SafeServiceException):
    pass


class InvalidInternalTx(InvalidMultisigTx):
    pass


class SignatureNotProvidedByOwner(InvalidMultisigTx):
    pass


class SignaturesDataTooShort(InvalidMultisigTx):
    pass


class InvalidOwnerProvided(InvalidMultisigTx):
    pass


class InvalidSignaturesProvided(InvalidMultisigTx):
    pass


class CouldNotPayGasWithEther(InvalidMultisigTx):
    pass


class CouldNotPayGasWithToken(InvalidMultisigTx):
    pass


class HashHasNotBeenApproved(InvalidMultisigTx):
    pass


class NotEnoughSafeTransactionGas(InvalidMultisigTx):
    pass


class InvalidContractSignatureLocation(InvalidMultisigTx):
    pass


class OnlyOwnersCanApproveAHash(InvalidMultisigTx):
    pass


class OwnerManagerException(InvalidMultisigTx):
    pass


class ModuleManagerException(InvalidMultisigTx):
    pass



import logging
from typing import List, Optional

from django.conf import settings

from eth_account import Account
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3.contract import Contract

from gnosis.eth.contracts import (get_multi_send_contract,
                                  get_proxy_factory_contract,
                                  get_safe_contract, get_safe_V1_0_0_contract,
                                  get_safe_V1_3_0_contract)
from gnosis.eth.tests.ethereum_test_case import EthereumTestCaseMixin
from gnosis.safe import Safe
from gnosis.safe.multi_send import MultiSend
from gnosis.safe.proxy_factory import ProxyFactory
from gnosis.safe.safe_create2_tx import SafeCreate2Tx

from ...eth.constants import NULL_ADDRESS
from .utils import generate_salt_nonce

logger = logging.getLogger(__name__)


contract_addresses = {
    'safe': Safe.deploy_master_contract,
    'safe_V0_0_1': Safe.deploy_old_master_contract,
    'safe_V1_0_0': Safe.deploy_master_contract_v1_0_0,
    'safe_V1_3_0': Safe.deploy_master_contract_v1_3_0,
    'proxy_factory': ProxyFactory.deploy_proxy_factory_contract,
    'proxy_factory_V1_0_0': ProxyFactory.deploy_proxy_factory_contract_v1_0_0,
    'multi_send': MultiSend.deploy_contract,
}


class SafeTestCaseMixin(EthereumTestCaseMixin):
    safe_contract_address: str
    safe_contract: Contract
    safe_contract_V1_3_0_address: str
    safe_contract_V1_3_0: Contract
    safe_contract_V1_0_0_address: str
    safe_contract_V1_0_0: Contract
    safe_contract_V0_0_1_address: str
    safe_contract_V0_0_1: Contract
    proxy_factory_contract_address: str
    proxy_factory_contract: Contract
    proxy_factory: ProxyFactory
    multi_send_contract: Contract
    multi_send: MultiSend

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        for key, value in contract_addresses.items():
            if callable(value):
                contract_addresses[key] = value(cls.ethereum_client, cls.ethereum_test_account).contract_address

        settings.SAFE_CONTRACT_ADDRESS = contract_addresses['safe']
        settings.SAFE_MULTISEND_ADDRESS = contract_addresses['multi_send']
        settings.SAFE_V1_3_0_CONTRACT_ADDRESS = contract_addresses['safe_V1_3_0']
        settings.SAFE_V1_0_0_CONTRACT_ADDRESS = contract_addresses['safe_V1_0_0']
        settings.SAFE_V0_0_1_CONTRACT_ADDRESS = contract_addresses['safe_V0_0_1']
        settings.SAFE_PROXY_FACTORY_ADDRESS = contract_addresses['proxy_factory']
        settings.SAFE_PROXY_FACTORY_V1_0_0_ADDRESS = contract_addresses['proxy_factory_V1_0_0']
        settings.SAFE_VALID_CONTRACT_ADDRESSES = {settings.SAFE_CONTRACT_ADDRESS,
                                                  settings.SAFE_V1_0_0_CONTRACT_ADDRESS,
                                                  settings.SAFE_V0_0_1_CONTRACT_ADDRESS,
                                                  }
        cls.safe_contract_address = contract_addresses['safe']
        cls.safe_contract = get_safe_contract(cls.w3, cls.safe_contract_address)
        cls.safe_contract_V1_3_0_address = contract_addresses['safe_V1_3_0']
        cls.safe_contract_V1_3_0 = get_safe_V1_3_0_contract(cls.w3, cls.safe_contract_V1_3_0_address)
        cls.safe_contract_V1_0_0_address = contract_addresses['safe_V1_0_0']
        cls.safe_contract_V1_0_0 = get_safe_V1_0_0_contract(cls.w3, cls.safe_contract_V1_0_0_address)
        cls.safe_contract_V0_0_1_address = contract_addresses['safe_V0_0_1']
        cls.safe_contract_V0_0_1 = get_safe_V1_0_0_contract(cls.w3, cls.safe_contract_V0_0_1_address)
        cls.proxy_factory_contract_address = contract_addresses['proxy_factory']
        cls.proxy_factory_contract = get_proxy_factory_contract(cls.w3, cls.proxy_factory_contract_address)
        cls.proxy_factory = ProxyFactory(cls.proxy_factory_contract_address, cls.ethereum_client)
        cls.multi_send_contract = get_multi_send_contract(cls.w3, contract_addresses['multi_send'])
        cls.multi_send = MultiSend(cls.multi_send_contract.address, cls.ethereum_client)

    def build_test_safe(self, number_owners: int = 3, threshold: Optional[int] = None,
                        owners: Optional[List[str]] = None, fallback_handler: Optional[str] = None) -> SafeCreate2Tx:
        salt_nonce = generate_salt_nonce()
        owners = owners if owners else [Account.create().address for _ in range(number_owners)]
        threshold = threshold if threshold else len(owners) - 1

        gas_price = self.ethereum_client.w3.eth.gas_price
        return Safe.build_safe_create2_tx(self.ethereum_client, self.safe_contract_address,
                                          self.proxy_factory_contract_address, salt_nonce, owners, threshold,
                                          fallback_handler=fallback_handler,
                                          gas_price=gas_price, payment_token=None, fixed_creation_cost=0)

    def deploy_test_safe(self, number_owners: int = 3, threshold: Optional[int] = None,
                         owners: Optional[List[str]] = None, initial_funding_wei: int = 0,
                         fallback_handler: Optional[str] = None) -> SafeCreate2Tx:
        owners = owners if owners else [Account.create().address for _ in range(number_owners)]
        if not threshold:
            threshold = len(owners) - 1 if len(owners) > 1 else 1
        safe_creation_tx = self.build_test_safe(threshold=threshold, owners=owners, fallback_handler=fallback_handler)

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(self.ethereum_test_account,
                                                                               self.safe_contract_address,
                                                                               safe_creation_tx.safe_setup_data,
                                                                               safe_creation_tx.salt_nonce)

        safe_address = ethereum_tx_sent.contract_address
        if initial_funding_wei:
            self.send_ether(safe_address, initial_funding_wei)

        safe_instance = get_safe_contract(self.w3, safe_address)

        self.assertEqual(safe_instance.functions.getThreshold().call(), threshold)
        self.assertEqual(safe_instance.functions.getOwners().call(), owners)
        self.assertEqual(safe_address, safe_creation_tx.safe_address)

        return safe_creation_tx

    def deploy_test_safe_v1_0_0(self, number_owners: int = 3, threshold: Optional[int] = None,
                                owners: Optional[List[ChecksumAddress]] = None, initial_funding_wei: int = 0) -> Safe:
        owners = owners if owners else [Account.create().address for _ in range(number_owners)]
        if not threshold:
            threshold = len(owners) - 1 if len(owners) > 1 else 1
        empty_parameters = {'gas': 1, 'gasPrice': 1}
        to = NULL_ADDRESS
        data = b''
        payment_token = NULL_ADDRESS
        payment = 0
        payment_receiver = NULL_ADDRESS
        initializer = HexBytes(
            self.safe_contract_V1_0_0.functions.setup(
                owners, threshold, to, data, payment_token,
                payment, payment_receiver
            ).buildTransaction(empty_parameters)['data']
        )
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract(
            self.ethereum_test_account, self.safe_contract_V1_0_0_address,
            initializer=initializer
        )
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        if initial_funding_wei:
            self.send_ether(safe.address, initial_funding_wei)

        self.assertEqual(safe.retrieve_threshold(), threshold)
        self.assertCountEqual(safe.retrieve_owners(), owners)

        return safe

    def deploy_test_safe_v1_3_0(self, number_owners: int = 3, threshold: Optional[int] = None,
                                owners: Optional[List[ChecksumAddress]] = None, initial_funding_wei: int = 0,
                                fallback_handler: ChecksumAddress = NULL_ADDRESS) -> Safe:
        owners = owners if owners else [Account.create().address for _ in range(number_owners)]
        if not threshold:
            threshold = len(owners) - 1 if len(owners) > 1 else 1
        empty_parameters = {'gas': 1, 'gasPrice': 1}
        to = NULL_ADDRESS
        data = b''
        payment_token = NULL_ADDRESS
        payment = 0
        payment_receiver = NULL_ADDRESS
        initializer = HexBytes(
            self.safe_contract_V1_3_0.functions.setup(
                owners, threshold, to, data, fallback_handler, payment_token,
                payment, payment_receiver
            ).buildTransaction(empty_parameters)['data']
        )
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract(
            self.ethereum_test_account, self.safe_contract_V1_3_0.address,
            initializer=initializer
        )
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        if initial_funding_wei:
            self.send_ether(safe.address, initial_funding_wei)

        self.assertEqual(safe.retrieve_threshold(), threshold)
        self.assertCountEqual(safe.retrieve_owners(), owners)

        return safe



from typing import Any, Dict, List, NoReturn, Optional, Tuple, Type

from eth_account import Account
from hexbytes import HexBytes
from packaging.version import Version
from web3.exceptions import BadFunctionCallOutput, ContractLogicError
from web3.types import BlockIdentifier, TxParams, Wei

from py_eth_sig_utils.eip712 import encode_typed_data

from gnosis.eth import EthereumClient
from gnosis.eth.constants import NULL_ADDRESS
from gnosis.eth.contracts import get_safe_contract

from .exceptions import (CouldNotPayGasWithEther, CouldNotPayGasWithToken,
                         HashHasNotBeenApproved,
                         InvalidContractSignatureLocation, InvalidInternalTx,
                         InvalidMultisigTx, InvalidOwnerProvided,
                         InvalidSignaturesProvided, ModuleManagerException,
                         NotEnoughSafeTransactionGas,
                         OnlyOwnersCanApproveAHash, OwnerManagerException,
                         SignatureNotProvidedByOwner, SignaturesDataTooShort)
from .safe_signature import SafeSignature
from .signatures import signature_to_bytes

try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property


class SafeTx:
    tx: TxParams  # If executed, `tx` is set
    tx_hash: bytes  # If executed, `tx_hash` is set

    def __init__(self,
                 ethereum_client: EthereumClient,
                 safe_address: str,
                 to: str,
                 value: int,
                 data: bytes,
                 operation: int,
                 safe_tx_gas: int,
                 base_gas: int,
                 gas_price: int,
                 gas_token: str,
                 refund_receiver: str,
                 signatures: bytes = b'',
                 safe_nonce: Optional[int] = None,
                 safe_version: str = None,
                 chain_id: Optional[int] = None):
        """
        :param ethereum_client:
        :param safe_address:
        :param to:
        :param value:
        :param data:
        :param operation:
        :param safe_tx_gas:
        :param base_gas:
        :param gas_price:
        :param gas_token:
        :param refund_receiver:
        :param signatures:
        :param safe_nonce: Current nonce of the Safe. If not provided, it will be retrieved from network
        :param safe_version: Safe version 1.0.0 renamed `baseGas` to `dataGas`. Safe version 1.3.0 added `chainId` to
        the `domainSeparator`. If not provided, it will be retrieved from network
        :param chain_id: Ethereum network chain_id is used in hash calculation for Safes >= 1.3.0. If not provided,
        it will be retrieved from the provided ethereum_client
        """

        assert isinstance(signatures, bytes), 'Signatures must be bytes'
        self.ethereum_client = ethereum_client
        self.safe_address = safe_address
        self.to = to or NULL_ADDRESS
        self.value = value
        self.data = HexBytes(data) if data else b''
        self.operation = operation
        self.safe_tx_gas = safe_tx_gas
        self.base_gas = base_gas
        self.gas_price = gas_price
        self.gas_token = gas_token or NULL_ADDRESS
        self.refund_receiver = refund_receiver or NULL_ADDRESS
        self.signatures = signatures
        self._safe_nonce = safe_nonce
        self._safe_version = safe_version
        self._chain_id = chain_id

    def __str__(self):
        return f'SafeTx - safe={self.safe_address} - to={self.to} - value={self.value} - data={self.data.hex()} - ' \
               f'operation={self.operation} - safe-tx-gas={self.safe_tx_gas} - base-gas={self.base_gas} - ' \
               f'gas-price={self.gas_price} - gas-token={self.gas_token} - refund-receiver={self.refund_receiver} - ' \
               f'signers = {self.signers}'

    @property
    def w3(self):
        return self.ethereum_client.w3

    @cached_property
    def contract(self):
        return get_safe_contract(self.w3, address=self.safe_address)

    @cached_property
    def chain_id(self) -> int:
        if self._chain_id is not None:
            return self._chain_id
        else:
            return self.w3.eth.chain_id

    @cached_property
    def safe_nonce(self) -> str:
        if self._safe_nonce is not None:
            return self._safe_nonce
        else:
            return self.contract.functions.nonce().call()

    @cached_property
    def safe_version(self) -> str:
        if self._safe_version is not None:
            return self._safe_version
        else:
            return self.contract.functions.VERSION().call()

    @property
    def safe_tx_hash(self) -> HexBytes:
        data = self.data.hex() if self.data else ''

        # Safes >= 1.0.0 Renamed `baseGas` to `dataGas`
        safe_version = Version(self.safe_version)
        base_gas_name = 'baseGas' if safe_version >= Version('1.0.0') else 'dataGas'

        structured_data = {
            'types': {
                'EIP712Domain': [
                    {'name': 'verifyingContract', 'type': 'address'},
                ],
                'SafeTx': [
                    {'name': 'to', 'type': 'address'},
                    {'name': 'value', 'type': 'uint256'},
                    {'name': 'data', 'type': 'bytes'},
                    {'name': 'operation', 'type': 'uint8'},
                    {'name': 'safeTxGas', 'type': 'uint256'},
                    {'name': base_gas_name, 'type': 'uint256'},
                    {'name': 'gasPrice', 'type': 'uint256'},
                    {'name': 'gasToken', 'type': 'address'},
                    {'name': 'refundReceiver', 'type': 'address'},
                    {'name': 'nonce', 'type': 'uint256'}
                ]
            },
            'primaryType': 'SafeTx',
            'domain': {
                'verifyingContract': self.safe_address,
            },
            'message': {
                'to': self.to,
                'value': self.value,
                'data': data,
                'operation': self.operation,
                'safeTxGas': self.safe_tx_gas,
                base_gas_name: self.base_gas,
                'gasPrice': self.gas_price,
                'gasToken': self.gas_token,
                'refundReceiver': self.refund_receiver,
                'nonce': self.safe_nonce,
            },
        }

        # Safes >= 1.3.1 Added `chainId` to the domain
        if safe_version >= Version('1.3.1'):
            # EIP712Domain(uint256 chainId,address verifyingContract)
            structured_data['types']['EIP712Domain'].insert(0, {'name': 'chainId', 'type': 'uint256'})
            structured_data['domain']['chainId'] = self.chain_id

        return HexBytes(encode_typed_data(structured_data))

    @property
    def signers(self) -> List[str]:
        return [safe_signature.owner for safe_signature
                in SafeSignature.parse_signature(self.signatures, self.safe_tx_hash)]

    @property
    def sorted_signers(self):
        return sorted(self.signers, key=lambda x: int(x, 16))

    @property
    def w3_tx(self):
        """
        :return: Web3 contract tx prepared for `call`, `transact` or `buildTransaction`
        """
        return self.contract.functions.execTransaction(
            self.to,
            self.value,
            self.data,
            self.operation,
            self.safe_tx_gas,
            self.base_gas,
            self.gas_price,
            self.gas_token,
            self.refund_receiver,
            self.signatures)

    def _raise_safe_vm_exception(self, message: str) -> NoReturn:
        error_with_exception: Dict[str, Type[InvalidMultisigTx]] = {
            'Could not pay gas costs with ether': CouldNotPayGasWithEther,
            'Could not pay gas costs with token': CouldNotPayGasWithToken,
            'Hash has not been approved': HashHasNotBeenApproved,
            'Hash not approved': HashHasNotBeenApproved,
            'Invalid contract signature location: data not complete': InvalidContractSignatureLocation,
            'Invalid contract signature location: inside static part': InvalidContractSignatureLocation,
            'Invalid contract signature location: length not present': InvalidContractSignatureLocation,
            'Invalid contract signature provided': InvalidContractSignatureLocation,
            'Invalid owner provided': InvalidOwnerProvided,
            'Invalid owner address provided': InvalidOwnerProvided,
            'Invalid signatures provided': InvalidSignaturesProvided,
            'Not enough gas to execute safe transaction': NotEnoughSafeTransactionGas,
            'Only owners can approve a hash': OnlyOwnersCanApproveAHash,
            'Signature not provided by owner': SignatureNotProvidedByOwner,
            'Signatures data too short': SignaturesDataTooShort,
            # OwnerManager
            'Address is already an owner': OwnerManagerException,
            'Invalid prevOwner, owner pair provided': OwnerManagerException,
            'New owner count needs to be larger than new threshold': OwnerManagerException,
            'Threshold cannot exceed owner count': OwnerManagerException,
            'Threshold needs to be greater than 0': OwnerManagerException,
            # ModuleManager
            'Invalid module address provided': ModuleManagerException,
            'Invalid prevModule, module pair provided': ModuleManagerException,
            'Method can only be called from an enabled module': ModuleManagerException,
            'Module has already been added': ModuleManagerException,
        }

        for reason, custom_exception in error_with_exception.items():
            if reason in message:
                raise custom_exception(message)
        raise InvalidMultisigTx(message)

    def call(self, tx_sender_address: Optional[str] = None, tx_gas: Optional[int] = None,
             block_identifier: Optional[BlockIdentifier] = 'latest') -> int:
        """
        :param tx_sender_address:
        :param tx_gas: Force a gas limit
        :param block_identifier:
        :return: `1` if everything ok
        """
        parameters: Dict[str, Any] = {
            'from': tx_sender_address if tx_sender_address else self.safe_address
        }

        if tx_gas:
            parameters['gas'] = tx_gas
        try:
            success = self.w3_tx.call(parameters, block_identifier=block_identifier)

            if not success:
                raise InvalidInternalTx('Success bit is %d, should be equal to 1' % success)
            return success
        except (ContractLogicError, BadFunctionCallOutput) as exc:
            # e.g. web3.exceptions.ContractLogicError: execution reverted: Invalid owner provided
            return self._raise_safe_vm_exception(str(exc))
        except ValueError as exc:  # Parity
            """
            Parity throws a ValueError, e.g.
            {'code': -32015,
             'message': 'VM execution error.',
             'data': 'Reverted 0x08c379a0000000000000000000000000000000000000000000000000000000000000020000000000000000
                      000000000000000000000000000000000000000000000001b496e76616c6964207369676e6174757265732070726f7669
                      6465640000000000'
            }
            """
            error_dict = exc.args[0]
            data = error_dict.get('data')
            if data and isinstance(data, str) and 'Reverted ' in data:
                # Parity
                result = HexBytes(data.replace('Reverted ', ''))
                return self._raise_safe_vm_exception(str(result))
            else:
                raise exc

    def recommended_gas(self) -> Wei:
        """
        :return: Recommended gas to use on the ethereum_tx
        """
        return Wei(self.base_gas + self.safe_tx_gas + 75000)

    def execute(self,
                tx_sender_private_key: str,
                tx_gas: Optional[int] = None,
                tx_gas_price: Optional[int] = None,
                tx_nonce: Optional[int] = None,
                block_identifier: Optional[BlockIdentifier] = 'latest') -> Tuple[HexBytes, TxParams]:
        """
        Send multisig tx to the Safe
        :param tx_sender_private_key: Sender private key
        :param tx_gas: Gas for the external tx. If not, `(safe_tx_gas + base_gas) * 2` will be used
        :param tx_gas_price: Gas price of the external tx. If not, `gas_price` will be used
        :param tx_nonce: Force nonce for `tx_sender`
        :param block_identifier: `latest` or `pending`
        :return: Tuple(tx_hash, tx)
        :raises: InvalidMultisigTx: If user tx cannot go through the Safe
        """

        sender_account = Account.from_key(tx_sender_private_key)
        tx_gas_price = tx_gas_price or self.gas_price or self.w3.eth.gas_price

        tx_parameters = {
            'from': sender_account.address,
            'gasPrice': tx_gas_price,
        }

        if tx_gas:
            tx_parameters['gas'] = tx_gas
        if tx_nonce is not None:
            tx_parameters['nonce'] = tx_nonce

        self.tx = self.w3_tx.buildTransaction(tx_parameters)
        self.tx['gas'] = Wei(tx_gas or (max(self.tx['gas'] + 75000, self.recommended_gas())))

        self.tx_hash = self.ethereum_client.send_unsigned_transaction(self.tx,
                                                                      private_key=sender_account.key,
                                                                      retry=False if tx_nonce is not None else True,
                                                                      block_identifier=block_identifier)

        # Set signatures empty after executing the tx. `Nonce` is increased even if it fails,
        # so signatures are not valid anymore
        self.signatures = b''
        return self.tx_hash, self.tx

    def sign(self, private_key: str) -> bytes:
        """
        {bytes32 r}{bytes32 s}{uint8 v}
        :param private_key:
        :return: Signature
        """
        account = Account.from_key(private_key)
        signature_dict = account.signHash(self.safe_tx_hash)
        signature = signature_to_bytes(signature_dict['v'],
                                       signature_dict['r'],
                                       signature_dict['s'])

        # Insert signature sorted
        if account.address not in self.signers:
            new_owners = self.signers + [account.address]
            new_owner_pos = sorted(new_owners,
                                   key=lambda x: int(x, 16)
                                   ).index(account.address)
            self.signatures = (self.signatures[: 65 * new_owner_pos] + signature
                               + self.signatures[65 * new_owner_pos:])
        return signature

    def unsign(self, address: str) -> bool:
        for pos, signer in enumerate(self.signers):
            if signer == address:
                self.signatures = self.signatures.replace(self.signatures[pos * 65: pos * 65 + 65], b'')
                return True
        return False



import logging

from django.test import TestCase

from eth_account import Account

from gnosis.eth.contracts import get_safe_contract, get_safe_V1_0_0_contract

from ..safe_create2_tx import SafeCreate2TxBuilder
from .safe_test_case import SafeTestCaseMixin
from .utils import generate_salt_nonce

logger = logging.getLogger(__name__)

LOG_TITLE_WIDTH = 100


class TestSafeCreationTx(SafeTestCaseMixin, TestCase):
    def test_safe_create2_tx_builder(self):
        w3 = self.w3

        salt_nonce = generate_salt_nonce()
        funder_account = self.ethereum_test_account
        owners = [Account.create().address for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price

        safe_creation_tx = SafeCreate2TxBuilder(w3=w3,
                                                master_copy_address=self.safe_contract_address,
                                                proxy_factory_address=self.proxy_factory_contract_address
                                                ).build(owners=owners,
                                                        threshold=threshold,
                                                        salt_nonce=salt_nonce,
                                                        gas_price=gas_price)

        self.assertEqual(safe_creation_tx.payment, safe_creation_tx.payment_ether)
        self.send_tx({
            'to': safe_creation_tx.safe_address,
            'value': safe_creation_tx.payment,
        }, funder_account)

        funder_balance = self.ethereum_client.get_balance(funder_account.address)
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(funder_account,
                                                                               self.safe_contract_address,
                                                                               safe_creation_tx.safe_setup_data,
                                                                               salt_nonce,
                                                                               safe_creation_tx.gas,
                                                                               safe_creation_tx.gas_price)
        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertEqual(tx_receipt.status, 1)

        # Funder balance must be bigger after a Safe deployment, as Safe deployment is a little overpriced
        self.assertGreater(self.ethereum_client.get_balance(funder_account.address), funder_balance)
        logs = self.proxy_factory_contract.events.ProxyCreation().processReceipt(tx_receipt)
        log = logs[0]
        self.assertIsNone(tx_receipt.contractAddress)
        self.assertEqual(log['event'], 'ProxyCreation')
        proxy_address = log['args']['proxy']
        self.assertEqual(proxy_address, safe_creation_tx.safe_address)
        self.assertEqual(ethereum_tx_sent.contract_address, safe_creation_tx.safe_address)

        deployed_safe_proxy_contract = get_safe_contract(w3, proxy_address)
        self.assertEqual(deployed_safe_proxy_contract.functions.getThreshold().call(), threshold)
        self.assertEqual(deployed_safe_proxy_contract.functions.getOwners().call(), owners)
        self.assertEqual(self.ethereum_client.get_balance(proxy_address), 0)

    def test_safe_create2_tx_builder_v_1_0_0(self):
        w3 = self.w3
        tx_hash = get_safe_V1_0_0_contract(self.w3).constructor().transact({
            'from': self.ethereum_test_account.address})
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        master_copy = tx_receipt['contractAddress']

        salt_nonce = generate_salt_nonce()
        funder_account = self.ethereum_test_account
        owners = [Account.create().address for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price

        safe_creation_tx = SafeCreate2TxBuilder(w3=w3,
                                                master_copy_address=master_copy,
                                                proxy_factory_address=self.proxy_factory_contract_address
                                                ).build(owners=owners,
                                                        threshold=threshold,
                                                        salt_nonce=salt_nonce,
                                                        gas_price=gas_price)

        self.assertEqual(safe_creation_tx.payment, safe_creation_tx.payment_ether)
        self.send_tx({
            'to': safe_creation_tx.safe_address,
            'value': safe_creation_tx.payment,
        }, funder_account)

        funder_balance = self.ethereum_client.get_balance(funder_account.address)
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(funder_account,
                                                                               master_copy,
                                                                               safe_creation_tx.safe_setup_data,
                                                                               salt_nonce,
                                                                               safe_creation_tx.gas,
                                                                               safe_creation_tx.gas_price)
        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertEqual(tx_receipt.status, 1)

        # Funder balance must be bigger after a Safe deployment, as Safe deployment is a little overpriced
        self.assertGreater(self.ethereum_client.get_balance(funder_account.address), funder_balance)
        logs = self.proxy_factory_contract.events.ProxyCreation().processReceipt(tx_receipt)
        log = logs[0]
        self.assertIsNone(tx_receipt.contractAddress)
        self.assertEqual(log['event'], 'ProxyCreation')
        proxy_address = log['args']['proxy']
        self.assertEqual(proxy_address, safe_creation_tx.safe_address)
        self.assertEqual(ethereum_tx_sent.contract_address, safe_creation_tx.safe_address)

        deployed_safe_proxy_contract = get_safe_contract(w3, proxy_address)
        self.assertEqual(deployed_safe_proxy_contract.functions.VERSION().call(), '1.0.0')
        self.assertEqual(deployed_safe_proxy_contract.functions.getThreshold().call(), threshold)
        self.assertEqual(deployed_safe_proxy_contract.functions.getOwners().call(), owners)
        self.assertEqual(self.ethereum_client.get_balance(proxy_address), 0)

    def test_safe_create2_tx_builder_with_payment_receiver(self):
        w3 = self.w3

        salt_nonce = generate_salt_nonce()
        payment_receiver = Account.create().address
        funder_account = self.ethereum_test_account
        owners = [Account.create().address for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price

        safe_creation_tx = SafeCreate2TxBuilder(w3=w3,
                                                master_copy_address=self.safe_contract_address,
                                                proxy_factory_address=self.proxy_factory_contract_address
                                                ).build(owners=owners,
                                                        threshold=threshold,
                                                        salt_nonce=salt_nonce,
                                                        gas_price=gas_price,
                                                        payment_receiver=payment_receiver)

        self.assertEqual(safe_creation_tx.payment, safe_creation_tx.payment_ether)
        self.send_tx({
            'to': safe_creation_tx.safe_address,
            'value': safe_creation_tx.payment,
        }, funder_account)

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(funder_account,
                                                                               self.safe_contract_address,
                                                                               safe_creation_tx.safe_setup_data,
                                                                               salt_nonce,
                                                                               gas=safe_creation_tx.gas,
                                                                               gas_price=safe_creation_tx.gas_price)
        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertEqual(tx_receipt.status, 1)
        logs = self.proxy_factory_contract.events.ProxyCreation().processReceipt(tx_receipt)
        log = logs[0]
        self.assertIsNone(tx_receipt.contractAddress)
        self.assertEqual(log['event'], 'ProxyCreation')
        proxy_address = log['args']['proxy']
        self.assertEqual(proxy_address, safe_creation_tx.safe_address)
        self.assertEqual(ethereum_tx_sent.contract_address, safe_creation_tx.safe_address)

        deployed_safe_proxy_contract = get_safe_contract(w3, proxy_address)
        self.assertEqual(deployed_safe_proxy_contract.functions.getThreshold().call(), threshold)
        self.assertEqual(deployed_safe_proxy_contract.functions.getOwners().call(), owners)
        self.assertEqual(self.ethereum_client.get_balance(proxy_address), 0)

        self.assertEqual(self.ethereum_client.get_balance(payment_receiver), safe_creation_tx.payment)

    def test_safe_create2_tx_builder_with_fixed_cost(self):
        w3 = self.w3

        salt_nonce = generate_salt_nonce()
        funder_account = self.ethereum_test_account
        owners = [Account.create().address for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price
        fixed_creation_cost = 123  # Wei

        safe_creation_tx = SafeCreate2TxBuilder(w3=w3,
                                                master_copy_address=self.safe_contract_address,
                                                proxy_factory_address=self.proxy_factory_contract_address
                                                ).build(owners=owners,
                                                        threshold=threshold,
                                                        salt_nonce=salt_nonce,
                                                        gas_price=gas_price,
                                                        fixed_creation_cost=fixed_creation_cost)

        self.assertEqual(safe_creation_tx.payment, fixed_creation_cost)
        self.assertEqual(safe_creation_tx.payment_ether, safe_creation_tx.gas * safe_creation_tx.gas_price)

        self.send_tx({
            'to': safe_creation_tx.safe_address,
            'value': safe_creation_tx.payment,
        }, funder_account)

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(funder_account,
                                                                               self.safe_contract_address,
                                                                               safe_creation_tx.safe_setup_data,
                                                                               salt_nonce,
                                                                               gas=safe_creation_tx.gas,
                                                                               gas_price=safe_creation_tx.gas_price)
        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertEqual(tx_receipt.status, 1)
        logs = self.proxy_factory_contract.events.ProxyCreation().processReceipt(tx_receipt)
        log = logs[0]
        self.assertIsNone(tx_receipt.contractAddress)
        self.assertEqual(log['event'], 'ProxyCreation')
        proxy_address = log['args']['proxy']
        self.assertEqual(proxy_address, safe_creation_tx.safe_address)
        self.assertEqual(ethereum_tx_sent.contract_address, safe_creation_tx.safe_address)

        deployed_safe_proxy_contract = get_safe_contract(w3, proxy_address)
        self.assertEqual(deployed_safe_proxy_contract.functions.getThreshold().call(), threshold)
        self.assertEqual(deployed_safe_proxy_contract.functions.getOwners().call(), owners)
        self.assertEqual(self.ethereum_client.get_balance(proxy_address), 0)

    def test_safe_create2_tx_builder_with_token_payment(self):
        w3 = self.w3

        salt_nonce = generate_salt_nonce()
        erc20_deployer = Account.create()
        funder_account = self.ethereum_test_account
        owners = [Account.create().address for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price
        token_amount = int(1e18)
        erc20_contract = self.deploy_example_erc20(token_amount, erc20_deployer.address)
        self.assertEqual(erc20_contract.functions.balanceOf(erc20_deployer.address).call(), token_amount)

        # Send something to the erc20 deployer
        self.send_tx({
            'to': erc20_deployer.address,
            'value': w3.toWei(1, 'ether')
        }, funder_account)

        safe_creation_tx = SafeCreate2TxBuilder(w3=w3,
                                                master_copy_address=self.safe_contract_address,
                                                proxy_factory_address=self.proxy_factory_contract_address
                                                ).build(owners=owners,
                                                        threshold=threshold,
                                                        salt_nonce=salt_nonce,
                                                        gas_price=gas_price,
                                                        payment_token=erc20_contract.address)

        self.assertEqual(safe_creation_tx.payment_token, erc20_contract.address)
        self.assertGreater(safe_creation_tx.payment, 0)
        self.assertEqual(safe_creation_tx.payment_ether, safe_creation_tx.gas * safe_creation_tx.gas_price)

        self.send_tx(erc20_contract.functions.transfer(safe_creation_tx.safe_address,
                                                       safe_creation_tx.payment).buildTransaction({'from': erc20_deployer.address}),
                     erc20_deployer)
        self.assertEqual(erc20_contract.functions.balanceOf(safe_creation_tx.safe_address).call(),
                         safe_creation_tx.payment)

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(funder_account,
                                                                               self.safe_contract_address,
                                                                               safe_creation_tx.safe_setup_data,
                                                                               salt_nonce,
                                                                               gas=safe_creation_tx.gas,
                                                                               gas_price=safe_creation_tx.gas_price)
        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertEqual(tx_receipt.status, 1)
        logs = self.proxy_factory_contract.events.ProxyCreation().processReceipt(tx_receipt)
        log = logs[0]
        self.assertIsNone(tx_receipt.contractAddress)
        self.assertEqual(log['event'], 'ProxyCreation')
        proxy_address = log['args']['proxy']
        self.assertEqual(proxy_address, safe_creation_tx.safe_address)
        self.assertEqual(ethereum_tx_sent.contract_address, safe_creation_tx.safe_address)

        deployed_safe_proxy_contract = get_safe_contract(w3, proxy_address)
        self.assertEqual(deployed_safe_proxy_contract.functions.getThreshold().call(), threshold)
        self.assertEqual(deployed_safe_proxy_contract.functions.getOwners().call(), owners)
        self.assertEqual(self.ethereum_client.get_balance(proxy_address), 0)

    def test_safe_gas_with_multiple_owners(self):
        logger.info("Test Safe Proxy create2 gas with multiple owners".center(LOG_TITLE_WIDTH, '-'))
        w3 = self.w3
        funder_account = self.ethereum_test_account
        number_of_accounts = 10
        for i in range(2, number_of_accounts):
            salt_nonce = generate_salt_nonce()
            owners = [Account.create().address for _ in range(i + 1)]
            threshold = len(owners) - 1
            gas_price = self.gas_price

            safe_creation_tx = SafeCreate2TxBuilder(w3=w3,
                                                    master_copy_address=self.safe_contract_address,
                                                    proxy_factory_address=self.proxy_factory_contract_address
                                                    ).build(owners=owners,
                                                            threshold=threshold,
                                                            salt_nonce=salt_nonce,
                                                            gas_price=gas_price)

            self.send_tx({
                'to': safe_creation_tx.safe_address,
                'value': safe_creation_tx.payment,
            }, funder_account)

            ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(funder_account,
                                                                                   self.safe_contract_address,
                                                                                   safe_creation_tx.safe_setup_data,
                                                                                   salt_nonce,
                                                                                   gas=safe_creation_tx.gas,
                                                                                   gas_price=safe_creation_tx.gas_price)
            tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
            self.assertEqual(tx_receipt.status, 1)
            logs = self.proxy_factory_contract.events.ProxyCreation().processReceipt(tx_receipt)
            log = logs[0]
            self.assertIsNone(tx_receipt.contractAddress)
            self.assertEqual(log['event'], 'ProxyCreation')
            proxy_address = log['args']['proxy']
            self.assertEqual(proxy_address, safe_creation_tx.safe_address)
            self.assertEqual(ethereum_tx_sent.contract_address, safe_creation_tx.safe_address)

            logger.info("Number of owners: %d - Gas estimated %d - Gas Used %d - Difference %d - Gas used per owner %d",
                        len(owners),
                        safe_creation_tx.gas,
                        tx_receipt.gasUsed,
                        safe_creation_tx.gas - tx_receipt.gasUsed,
                        tx_receipt.gasUsed // len(owners))



import logging

from django.utils.translation import gettext_lazy as _

from ethereum.utils import checksum_encode
from hexbytes import HexBytes
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from ..constants import (SIGNATURE_R_MAX_VALUE, SIGNATURE_R_MIN_VALUE,
                         SIGNATURE_S_MAX_VALUE, SIGNATURE_S_MIN_VALUE,
                         SIGNATURE_V_MAX_VALUE, SIGNATURE_V_MIN_VALUE)

logger = logging.getLogger(__name__)


# ================================================ #
#                Custom Fields
# ================================================ #
class EthereumAddressField(serializers.Field):
    """
    Ethereum address checksumed
    https://github.com/ethereum/EIPs/blob/master/EIPS/eip-55.md
    """

    def __init__(self, allow_zero_address: bool = False, allow_sentinel_address: bool = False, **kwargs):
        self.allow_zero_address = allow_zero_address
        self.allow_sentinel_address = allow_sentinel_address
        super().__init__(**kwargs)

    def to_representation(self, obj):
        return obj

    def to_internal_value(self, data):
        # Check if address is valid
        try:
            if checksum_encode(data) != data:
                raise ValueError
            elif int(data, 16) == 0 and not self.allow_zero_address:
                raise ValidationError("0x0 address is not allowed")
            elif int(data, 16) == 1 and not self.allow_sentinel_address:
                raise ValidationError("0x1 address is not allowed")
        except ValueError:
            raise ValidationError("Address %s is not checksumed" % data)
        except Exception:
            raise ValidationError("Address %s is not valid" % data)

        return data


class HexadecimalField(serializers.Field):
    """
    Serializes hexadecimal values starting by `0x`. Empty values should be None or just `0x`.
    """

    default_error_messages = {
        'invalid': _('{value} is not an hexadecimal value.'),
        'blank': _('This field may not be blank.'),
        'max_length': _('Ensure this field has no more than {max_length} hexadecimal chars (not counting 0x).'),
        'min_length': _('Ensure this field has at least {min_length} hexadecimal chars (not counting 0x).'),
    }

    def __init__(self, **kwargs):
        self.allow_blank = kwargs.pop('allow_blank', False)
        self.max_length = kwargs.pop('max_length', None)
        self.min_length = kwargs.pop('min_length', None)
        super().__init__(**kwargs)

    def to_representation(self, obj):
        if not obj:
            return '0x'

        # We can get another types like `memoryview` from django models. `to_internal_value` is not used
        # when you provide an object instead of a json using `data`. Make sure everything is HexBytes.
        if hasattr(obj, 'hex'):
            obj = HexBytes(obj.hex())
        elif not isinstance(obj, HexBytes):
            obj = HexBytes(obj)
        return obj.hex()

    def to_internal_value(self, data):
        if isinstance(data, (bytes, memoryview)):
            data = data.hex()
        elif isinstance(data, str):
            data = data.strip()  # Trim spaces
            if data.startswith('0x'):  # Remove 0x prefix
                data = data[2:]
        elif data is None:
            pass
        else:
            self.fail('invalid', value=data)

        if not data:
            if self.allow_blank:
                return None
            else:
                self.fail('blank')

        try:
            data_hex = HexBytes(data)
            data_len = len(data_hex)
            if self.min_length and data_len < self.min_length:
                self.fail('min_length', min_length=data_len)
            elif self.max_length and data_len > self.max_length:
                self.fail('max_length', max_length=data_len)
            return data_hex
        except ValueError:
            self.fail('invalid', value=data)


class Sha3HashField(HexadecimalField):
    def __init__(self, **kwargs):
        kwargs['max_length'] = 32
        kwargs['min_length'] = 32
        super().__init__(**kwargs)


# ================================================ #
#                Base Serializers
# ================================================ #
class SignatureSerializer(serializers.Serializer):
    v = serializers.IntegerField(min_value=SIGNATURE_V_MIN_VALUE,
                                 max_value=SIGNATURE_V_MAX_VALUE)
    r = serializers.IntegerField(min_value=SIGNATURE_R_MIN_VALUE,
                                 max_value=SIGNATURE_R_MAX_VALUE)
    s = serializers.IntegerField(min_value=SIGNATURE_S_MIN_VALUE,
                                 max_value=SIGNATURE_S_MAX_VALUE)


class TransactionSerializer(serializers.Serializer):
    from_ = EthereumAddressField()
    value = serializers.IntegerField(min_value=0)
    data = HexadecimalField()
    gas = serializers.IntegerField(min_value=0)
    gas_price = serializers.IntegerField(min_value=0)
    nonce = serializers.IntegerField(min_value=0)

    def get_fields(self):
        result = super().get_fields()
        # Rename `from_` to `from`
        from_ = result.pop('from_')
        result['from'] = from_
        return result


class TransactionResponseSerializer(serializers.Serializer):
    """
    Use chars to avoid problems with big ints (i.e. JavaScript)
    """
    from_ = EthereumAddressField()
    value = serializers.IntegerField(min_value=0)
    data = serializers.CharField()
    gas = serializers.CharField()
    gas_price = serializers.CharField()
    nonce = serializers.IntegerField(min_value=0)

    def get_fields(self):
        result = super().get_fields()
        # Rename `from_` to `from`
        from_ = result.pop('from_')
        result['from'] = from_
        return result



from django.core.exceptions import ValidationError
from django.forms import CharField as CharFieldForm
from django.utils.translation import gettext_lazy as _

import django_filters
from web3 import Web3


class EthereumAddressFieldForm(CharFieldForm):
    default_error_messages = {
        'invalid': _('Enter a valid checksummed Ethereum Address.'),
    }

    def prepare_value(self, value):
        return value

    def to_python(self, value):
        value = super().to_python(value)
        if value in self.empty_values:
            return None
        elif not Web3.isChecksumAddress(value):
            raise ValidationError(self.error_messages['invalid'], code='invalid')
        return value


class EthereumAddressFilter(django_filters.Filter):
    field_class = EthereumAddressFieldForm



curve_address_provider_abi = [{'name': 'NewAddressIdentifier',
                               'inputs': [{'type': 'uint256', 'name': 'id', 'indexed': True},
                                          {'type': 'address', 'name': 'addr', 'indexed': False},
                                          {'type': 'string', 'name': 'description', 'indexed': False}],
                               'anonymous': False,
                               'type': 'event'},
                              {'name': 'AddressModified',
                               'inputs': [{'type': 'uint256', 'name': 'id', 'indexed': True},
                                          {'type': 'address', 'name': 'new_address', 'indexed': False},
                                          {'type': 'uint256', 'name': 'version', 'indexed': False}],
                               'anonymous': False,
                               'type': 'event'},
                              {'name': 'CommitNewAdmin',
                               'inputs': [{'type': 'uint256', 'name': 'deadline', 'indexed': True},
                                          {'type': 'address', 'name': 'admin', 'indexed': True}],
                               'anonymous': False,
                               'type': 'event'},
                              {'name': 'NewAdmin',
                               'inputs': [{'type': 'address', 'name': 'admin', 'indexed': True}],
                               'anonymous': False,
                               'type': 'event'},
                              {'outputs': [],
                               'inputs': [{'type': 'address', 'name': '_admin'}],
                               'stateMutability': 'nonpayable',
                               'type': 'constructor'},
                              {'name': 'get_registry',
                               'outputs': [{'type': 'address', 'name': ''}],
                               'inputs': [],
                               'stateMutability': 'view',
                               'type': 'function',
                               'gas': 1061},
                              {'name': 'max_id',
                               'outputs': [{'type': 'uint256', 'name': ''}],
                               'inputs': [],
                               'stateMutability': 'view',
                               'type': 'function',
                               'gas': 1258},
                              {'name': 'get_address',
                               'outputs': [{'type': 'address', 'name': ''}],
                               'inputs': [{'type': 'uint256', 'name': '_id'}],
                               'stateMutability': 'view',
                               'type': 'function',
                               'gas': 1308},
                              {'name': 'add_new_id',
                               'outputs': [{'type': 'uint256', 'name': ''}],
                               'inputs': [{'type': 'address', 'name': '_address'},
                                          {'type': 'string', 'name': '_description'}],
                               'stateMutability': 'nonpayable',
                               'type': 'function',
                               'gas': 291275},
                              {'name': 'set_address',
                               'outputs': [{'type': 'bool', 'name': ''}],
                               'inputs': [{'type': 'uint256', 'name': '_id'},
                                          {'type': 'address', 'name': '_address'}],
                               'stateMutability': 'nonpayable',
                               'type': 'function',
                               'gas': 182430},
                              {'name': 'unset_address',
                               'outputs': [{'type': 'bool', 'name': ''}],
                               'inputs': [{'type': 'uint256', 'name': '_id'}],
                               'stateMutability': 'nonpayable',
                               'type': 'function',
                               'gas': 101348},
                              {'name': 'commit_transfer_ownership',
                               'outputs': [{'type': 'bool', 'name': ''}],
                               'inputs': [{'type': 'address', 'name': '_new_admin'}],
                               'stateMutability': 'nonpayable',
                               'type': 'function',
                               'gas': 74048},
                              {'name': 'apply_transfer_ownership',
                               'outputs': [{'type': 'bool', 'name': ''}],
                               'inputs': [],
                               'stateMutability': 'nonpayable',
                               'type': 'function',
                               'gas': 60125},
                              {'name': 'revert_transfer_ownership',
                               'outputs': [{'type': 'bool', 'name': ''}],
                               'inputs': [],
                               'stateMutability': 'nonpayable',
                               'type': 'function',
                               'gas': 21400},
                              {'name': 'admin',
                               'outputs': [{'type': 'address', 'name': ''}],
                               'inputs': [],
                               'stateMutability': 'view',
                               'type': 'function',
                               'gas': 1331},
                              {'name': 'transfer_ownership_deadline',
                               'outputs': [{'type': 'uint256', 'name': ''}],
                               'inputs': [],
                               'stateMutability': 'view',
                               'type': 'function',
                               'gas': 1361},
                              {'name': 'future_admin',
                               'outputs': [{'type': 'address', 'name': ''}],
                               'inputs': [],
                               'stateMutability': 'view',
                               'type': 'function',
                               'gas': 1391},
                              {'name': 'get_id_info',
                               'outputs': [{'type': 'address', 'name': 'addr'},
                                           {'type': 'bool', 'name': 'is_active'},
                                           {'type': 'uint256', 'name': 'version'},
                                           {'type': 'uint256', 'name': 'last_modified'},
                                           {'type': 'string', 'name': 'description'}],
                               'inputs': [{'type': 'uint256', 'name': 'arg0'}],
                               'stateMutability': 'view',
                               'type': 'function',
                               'gas': 12168}]

curve_registry_abi = [{'name': 'PoolAdded',
                       'inputs': [{'type': 'address', 'name': 'pool', 'indexed': True},
                                  {'type': 'bytes', 'name': 'rate_method_id', 'indexed': False}],
                       'anonymous': False,
                       'type': 'event'},
                      {'name': 'PoolRemoved',
                       'inputs': [{'type': 'address', 'name': 'pool', 'indexed': True}],
                       'anonymous': False,
                       'type': 'event'},
                      {'outputs': [],
                       'inputs': [{'type': 'address', 'name': '_address_provider'},
                                  {'type': 'address', 'name': '_gauge_controller'}],
                       'stateMutability': 'nonpayable',
                       'type': 'constructor'},
                      {'name': 'find_pool_for_coins',
                       'outputs': [{'type': 'address', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_from'},
                                  {'type': 'address', 'name': '_to'}],
                       'stateMutability': 'view',
                       'type': 'function'},
                      {'name': 'find_pool_for_coins',
                       'outputs': [{'type': 'address', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_from'},
                                  {'type': 'address', 'name': '_to'},
                                  {'type': 'uint256', 'name': 'i'}],
                       'stateMutability': 'view',
                       'type': 'function'},
                      {'name': 'get_n_coins',
                       'outputs': [{'type': 'uint256[2]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 1704},
                      {'name': 'get_coins',
                       'outputs': [{'type': 'address[8]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 12285},
                      {'name': 'get_underlying_coins',
                       'outputs': [{'type': 'address[8]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 12347},
                      {'name': 'get_decimals',
                       'outputs': [{'type': 'uint256[8]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 8199},
                      {'name': 'get_underlying_decimals',
                       'outputs': [{'type': 'uint256[8]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 8261},
                      {'name': 'get_rates',
                       'outputs': [{'type': 'uint256[8]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 34780},
                      {'name': 'get_gauges',
                       'outputs': [{'type': 'address[10]', 'name': ''},
                                   {'type': 'int128[10]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 20310},
                      {'name': 'get_balances',
                       'outputs': [{'type': 'uint256[8]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 16818},
                      {'name': 'get_underlying_balances',
                       'outputs': [{'type': 'uint256[8]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 158953},
                      {'name': 'get_virtual_price_from_lp_token',
                       'outputs': [{'type': 'uint256', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_token'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 2080},
                      {'name': 'get_A',
                       'outputs': [{'type': 'uint256', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 1198},
                      {'name': 'get_parameters',
                       'outputs': [{'type': 'uint256', 'name': 'A'},
                                   {'type': 'uint256', 'name': 'future_A'},
                                   {'type': 'uint256', 'name': 'fee'},
                                   {'type': 'uint256', 'name': 'admin_fee'},
                                   {'type': 'uint256', 'name': 'future_fee'},
                                   {'type': 'uint256', 'name': 'future_admin_fee'},
                                   {'type': 'address', 'name': 'future_owner'},
                                   {'type': 'uint256', 'name': 'initial_A'},
                                   {'type': 'uint256', 'name': 'initial_A_time'},
                                   {'type': 'uint256', 'name': 'future_A_time'}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 6458},
                      {'name': 'get_fees',
                       'outputs': [{'type': 'uint256[2]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 1603},
                      {'name': 'get_admin_balances',
                       'outputs': [{'type': 'uint256[8]', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 36719},
                      {'name': 'get_coin_indices',
                       'outputs': [{'type': 'int128', 'name': ''},
                                   {'type': 'int128', 'name': ''},
                                   {'type': 'bool', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'},
                                  {'type': 'address', 'name': '_from'},
                                  {'type': 'address', 'name': '_to'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 27456},
                      {'name': 'estimate_gas_used',
                       'outputs': [{'type': 'uint256', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': '_pool'},
                                  {'type': 'address', 'name': '_from'},
                                  {'type': 'address', 'name': '_to'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 32329},
                      {'name': 'add_pool',
                       'outputs': [],
                       'inputs': [{'type': 'address', 'name': '_pool'},
                                  {'type': 'uint256', 'name': '_n_coins'},
                                  {'type': 'address', 'name': '_lp_token'},
                                  {'type': 'bytes32', 'name': '_rate_method_id'},
                                  {'type': 'uint256', 'name': '_decimals'},
                                  {'type': 'uint256', 'name': '_underlying_decimals'},
                                  {'type': 'bool', 'name': '_has_initial_A'},
                                  {'type': 'bool', 'name': '_is_v1'}],
                       'stateMutability': 'nonpayable',
                       'type': 'function',
                       'gas': 10196577},
                      {'name': 'add_pool_without_underlying',
                       'outputs': [],
                       'inputs': [{'type': 'address', 'name': '_pool'},
                                  {'type': 'uint256', 'name': '_n_coins'},
                                  {'type': 'address', 'name': '_lp_token'},
                                  {'type': 'bytes32', 'name': '_rate_method_id'},
                                  {'type': 'uint256', 'name': '_decimals'},
                                  {'type': 'uint256', 'name': '_use_rates'},
                                  {'type': 'bool', 'name': '_has_initial_A'},
                                  {'type': 'bool', 'name': '_is_v1'}],
                       'stateMutability': 'nonpayable',
                       'type': 'function',
                       'gas': 5590664},
                      {'name': 'add_metapool',
                       'outputs': [],
                       'inputs': [{'type': 'address', 'name': '_pool'},
                                  {'type': 'uint256', 'name': '_n_coins'},
                                  {'type': 'address', 'name': '_lp_token'},
                                  {'type': 'uint256', 'name': '_decimals'}],
                       'stateMutability': 'nonpayable',
                       'type': 'function',
                       'gas': 10226976},
                      {'name': 'remove_pool',
                       'outputs': [],
                       'inputs': [{'type': 'address', 'name': '_pool'}],
                       'stateMutability': 'nonpayable',
                       'type': 'function',
                       'gas': 779646579509},
                      {'name': 'set_pool_gas_estimates',
                       'outputs': [],
                       'inputs': [{'type': 'address[5]', 'name': '_addr'},
                                  {'type': 'uint256[2][5]', 'name': '_amount'}],
                       'stateMutability': 'nonpayable',
                       'type': 'function',
                       'gas': 355578},
                      {'name': 'set_coin_gas_estimates',
                       'outputs': [],
                       'inputs': [{'type': 'address[10]', 'name': '_addr'},
                                  {'type': 'uint256[10]', 'name': '_amount'}],
                       'stateMutability': 'nonpayable',
                       'type': 'function',
                       'gas': 357165},
                      {'name': 'set_gas_estimate_contract',
                       'outputs': [],
                       'inputs': [{'type': 'address', 'name': '_pool'},
                                  {'type': 'address', 'name': '_estimator'}],
                       'stateMutability': 'nonpayable',
                       'type': 'function',
                       'gas': 37747},
                      {'name': 'set_liquidity_gauges',
                       'outputs': [],
                       'inputs': [{'type': 'address', 'name': '_pool'},
                                  {'type': 'address[10]', 'name': '_liquidity_gauges'}],
                       'stateMutability': 'nonpayable',
                       'type': 'function',
                       'gas': 365793},
                      {'name': 'address_provider',
                       'outputs': [{'type': 'address', 'name': ''}],
                       'inputs': [],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 2111},
                      {'name': 'gauge_controller',
                       'outputs': [{'type': 'address', 'name': ''}],
                       'inputs': [],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 2141},
                      {'name': 'pool_list',
                       'outputs': [{'type': 'address', 'name': ''}],
                       'inputs': [{'type': 'uint256', 'name': 'arg0'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 2280},
                      {'name': 'pool_count',
                       'outputs': [{'type': 'uint256', 'name': ''}],
                       'inputs': [],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 2201},
                      {'name': 'get_pool_from_lp_token',
                       'outputs': [{'type': 'address', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': 'arg0'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 2446},
                      {'name': 'get_lp_token',
                       'outputs': [{'type': 'address', 'name': ''}],
                       'inputs': [{'type': 'address', 'name': 'arg0'}],
                       'stateMutability': 'view',
                       'type': 'function',
                       'gas': 2476}]

curve_pool_abi = [{'name': 'TokenExchange',
                   'inputs': [{'type': 'address', 'name': 'buyer', 'indexed': True},
                              {'type': 'int128', 'name': 'sold_id', 'indexed': False},
                              {'type': 'uint256', 'name': 'tokens_sold', 'indexed': False},
                              {'type': 'int128', 'name': 'bought_id', 'indexed': False},
                              {'type': 'uint256', 'name': 'tokens_bought', 'indexed': False}],
                   'anonymous': False,
                   'type': 'event'},
                  {'name': 'TokenExchangeUnderlying',
                   'inputs': [{'type': 'address', 'name': 'buyer', 'indexed': True},
                              {'type': 'int128', 'name': 'sold_id', 'indexed': False},
                              {'type': 'uint256', 'name': 'tokens_sold', 'indexed': False},
                              {'type': 'int128', 'name': 'bought_id', 'indexed': False},
                              {'type': 'uint256', 'name': 'tokens_bought', 'indexed': False}],
                   'anonymous': False,
                   'type': 'event'},
                  {'name': 'AddLiquidity',
                   'inputs': [{'type': 'address', 'name': 'provider', 'indexed': True},
                              {'type': 'uint256[4]', 'name': 'token_amounts', 'indexed': False},
                              {'type': 'uint256[4]', 'name': 'fees', 'indexed': False},
                              {'type': 'uint256', 'name': 'invariant', 'indexed': False},
                              {'type': 'uint256', 'name': 'token_supply', 'indexed': False}],
                   'anonymous': False,
                   'type': 'event'},
                  {'name': 'RemoveLiquidity',
                   'inputs': [{'type': 'address', 'name': 'provider', 'indexed': True},
                              {'type': 'uint256[4]', 'name': 'token_amounts', 'indexed': False},
                              {'type': 'uint256[4]', 'name': 'fees', 'indexed': False},
                              {'type': 'uint256', 'name': 'token_supply', 'indexed': False}],
                   'anonymous': False,
                   'type': 'event'},
                  {'name': 'RemoveLiquidityImbalance',
                   'inputs': [{'type': 'address', 'name': 'provider', 'indexed': True},
                              {'type': 'uint256[4]', 'name': 'token_amounts', 'indexed': False},
                              {'type': 'uint256[4]', 'name': 'fees', 'indexed': False},
                              {'type': 'uint256', 'name': 'invariant', 'indexed': False},
                              {'type': 'uint256', 'name': 'token_supply', 'indexed': False}],
                   'anonymous': False,
                   'type': 'event'},
                  {'name': 'CommitNewAdmin',
                   'inputs': [{'type': 'uint256',
                               'name': 'deadline',
                               'indexed': True,
                               'unit': 'sec'},
                              {'type': 'address', 'name': 'admin', 'indexed': True}],
                   'anonymous': False,
                   'type': 'event'},
                  {'name': 'NewAdmin',
                   'inputs': [{'type': 'address', 'name': 'admin', 'indexed': True}],
                   'anonymous': False,
                   'type': 'event'},
                  {'name': 'CommitNewParameters',
                   'inputs': [{'type': 'uint256',
                               'name': 'deadline',
                               'indexed': True,
                               'unit': 'sec'},
                              {'type': 'uint256', 'name': 'A', 'indexed': False},
                              {'type': 'uint256', 'name': 'fee', 'indexed': False},
                              {'type': 'uint256', 'name': 'admin_fee', 'indexed': False}],
                   'anonymous': False,
                   'type': 'event'},
                  {'name': 'NewParameters',
                   'inputs': [{'type': 'uint256', 'name': 'A', 'indexed': False},
                              {'type': 'uint256', 'name': 'fee', 'indexed': False},
                              {'type': 'uint256', 'name': 'admin_fee', 'indexed': False}],
                   'anonymous': False,
                   'type': 'event'},
                  {'outputs': [],
                   'inputs': [{'type': 'address[4]', 'name': '_coins'},
                              {'type': 'address[4]', 'name': '_underlying_coins'},
                              {'type': 'address', 'name': '_pool_token'},
                              {'type': 'uint256', 'name': '_A'},
                              {'type': 'uint256', 'name': '_fee'}],
                   'constant': False,
                   'payable': False,
                   'type': 'constructor'},
                  {'name': 'get_virtual_price',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 1570535},
                  {'name': 'calc_token_amount',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [{'type': 'uint256[4]', 'name': 'amounts'},
                              {'type': 'bool', 'name': 'deposit'}],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 6103471},
                  {'name': 'add_liquidity',
                   'outputs': [],
                   'inputs': [{'type': 'uint256[4]', 'name': 'amounts'},
                              {'type': 'uint256', 'name': 'min_mint_amount'}],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 9331701},
                  {'name': 'get_dy',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [{'type': 'int128', 'name': 'i'},
                              {'type': 'int128', 'name': 'j'},
                              {'type': 'uint256', 'name': 'dx'}],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 3489637},
                  {'name': 'get_dy_underlying',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [{'type': 'int128', 'name': 'i'},
                              {'type': 'int128', 'name': 'j'},
                              {'type': 'uint256', 'name': 'dx'}],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 3489467},
                  {'name': 'exchange',
                   'outputs': [],
                   'inputs': [{'type': 'int128', 'name': 'i'},
                              {'type': 'int128', 'name': 'j'},
                              {'type': 'uint256', 'name': 'dx'},
                              {'type': 'uint256', 'name': 'min_dy'}],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 7034253},
                  {'name': 'exchange_underlying',
                   'outputs': [],
                   'inputs': [{'type': 'int128', 'name': 'i'},
                              {'type': 'int128', 'name': 'j'},
                              {'type': 'uint256', 'name': 'dx'},
                              {'type': 'uint256', 'name': 'min_dy'}],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 7050488},
                  {'name': 'remove_liquidity',
                   'outputs': [],
                   'inputs': [{'type': 'uint256', 'name': '_amount'},
                              {'type': 'uint256[4]', 'name': 'min_amounts'}],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 241191},
                  {'name': 'remove_liquidity_imbalance',
                   'outputs': [],
                   'inputs': [{'type': 'uint256[4]', 'name': 'amounts'},
                              {'type': 'uint256', 'name': 'max_burn_amount'}],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 9330864},
                  {'name': 'commit_new_parameters',
                   'outputs': [],
                   'inputs': [{'type': 'uint256', 'name': 'amplification'},
                              {'type': 'uint256', 'name': 'new_fee'},
                              {'type': 'uint256', 'name': 'new_admin_fee'}],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 146045},
                  {'name': 'apply_new_parameters',
                   'outputs': [],
                   'inputs': [],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 133452},
                  {'name': 'revert_new_parameters',
                   'outputs': [],
                   'inputs': [],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 21775},
                  {'name': 'commit_transfer_ownership',
                   'outputs': [],
                   'inputs': [{'type': 'address', 'name': '_owner'}],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 74452},
                  {'name': 'apply_transfer_ownership',
                   'outputs': [],
                   'inputs': [],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 60508},
                  {'name': 'revert_transfer_ownership',
                   'outputs': [],
                   'inputs': [],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 21865},
                  {'name': 'withdraw_admin_fees',
                   'outputs': [],
                   'inputs': [],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 23448},
                  {'name': 'kill_me',
                   'outputs': [],
                   'inputs': [],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 37818},
                  {'name': 'unkill_me',
                   'outputs': [],
                   'inputs': [],
                   'constant': False,
                   'payable': False,
                   'type': 'function',
                   'gas': 21955},
                  {'name': 'coins',
                   'outputs': [{'type': 'address', 'name': ''}],
                   'inputs': [{'type': 'int128', 'name': 'arg0'}],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2130},
                  {'name': 'underlying_coins',
                   'outputs': [{'type': 'address', 'name': ''}],
                   'inputs': [{'type': 'int128', 'name': 'arg0'}],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2160},
                  {'name': 'balances',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [{'type': 'int128', 'name': 'arg0'}],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2190},
                  {'name': 'A',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2021},
                  {'name': 'fee',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2051},
                  {'name': 'admin_fee',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2081},
                  {'name': 'owner',
                   'outputs': [{'type': 'address', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2111},
                  {'name': 'admin_actions_deadline',
                   'outputs': [{'type': 'uint256', 'unit': 'sec', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2141},
                  {'name': 'transfer_ownership_deadline',
                   'outputs': [{'type': 'uint256', 'unit': 'sec', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2171},
                  {'name': 'future_A',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2201},
                  {'name': 'future_fee',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2231},
                  {'name': 'future_admin_fee',
                   'outputs': [{'type': 'uint256', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2261},
                  {'name': 'future_owner',
                   'outputs': [{'type': 'address', 'name': ''}],
                   'inputs': [],
                   'constant': True,
                   'payable': False,
                   'type': 'function',
                   'gas': 2291}]



from logging import getLogger
from typing import Optional

from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract import Contract

from gnosis.eth import EthereumClient
from gnosis.eth.contracts import (get_paying_proxy_deployed_bytecode,
                                  get_proxy_1_0_0_deployed_bytecode,
                                  get_proxy_factory_contract,
                                  get_proxy_factory_V1_0_0_contract)
from gnosis.eth.ethereum_client import EthereumTxSent
from gnosis.eth.utils import compare_byte_code

logger = getLogger(__name__)


class ProxyFactory:
    proxy_runtime_code: Optional[bytes] = None  # Cache runtime code

    def __init__(self, address: str, ethereum_client: EthereumClient):
        assert Web3.isChecksumAddress(address), \
            '%s proxy factory address not valid' % address

        self.address = address
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3

    @staticmethod
    def _deploy_proxy_factory_contract(ethereum_client: EthereumClient,
                                       deployer_account: LocalAccount, contract: Contract) -> EthereumTxSent:
        tx = contract.constructor().buildTransaction({'from': deployer_account.address})

        tx_hash = ethereum_client.send_unsigned_transaction(tx, private_key=deployer_account.key)
        tx_receipt = ethereum_client.get_transaction_receipt(tx_hash, timeout=120)
        assert tx_receipt
        assert tx_receipt['status']
        contract_address = tx_receipt['contractAddress']
        logger.info("Deployed and initialized Proxy Factory Contract=%s by %s", contract_address,
                    deployer_account.address)
        return EthereumTxSent(tx_hash, tx, contract_address)

    @classmethod
    def deploy_proxy_factory_contract(cls, ethereum_client: EthereumClient,
                                      deployer_account: LocalAccount) -> EthereumTxSent:
        """
        Deploy proxy factory contract
        :param ethereum_client:
        :param deployer_account: Ethereum Account
        :return: deployed contract address
        """
        proxy_factory_contract = get_proxy_factory_contract(ethereum_client.w3)
        return cls._deploy_proxy_factory_contract(ethereum_client, deployer_account, proxy_factory_contract)

    @classmethod
    def deploy_proxy_factory_contract_v1_0_0(cls, ethereum_client: EthereumClient,
                                             deployer_account: LocalAccount) -> EthereumTxSent:
        """
        Deploy proxy factory contract
        :param ethereum_client:
        :param deployer_account: Ethereum Account
        :return: deployed contract address
        """
        proxy_factory_contract = get_proxy_factory_V1_0_0_contract(ethereum_client.w3)
        return cls._deploy_proxy_factory_contract(ethereum_client, deployer_account, proxy_factory_contract)

    def check_proxy_code(self, address: ChecksumAddress) -> bool:
        """
        Check if proxy is valid
        :param address: Ethereum address to check
        :return: True if proxy is valid, False otherwise
        """

        deployed_proxy_code = self.w3.eth.get_code(address)
        proxy_code_fns = (get_paying_proxy_deployed_bytecode,
                          get_proxy_1_0_0_deployed_bytecode,
                          self.get_proxy_runtime_code)
        for proxy_code_fn in proxy_code_fns:
            if compare_byte_code(deployed_proxy_code, proxy_code_fn()):
                return True
        return False

    def deploy_proxy_contract(self, deployer_account: LocalAccount, master_copy: str, initializer: bytes = b'',
                              gas: Optional[int] = None, gas_price: Optional[int] = None) -> EthereumTxSent:
        """
        Deploy proxy contract via ProxyFactory using `createProxy` function
        :param deployer_account: Ethereum account
        :param master_copy: Address the proxy will point at
        :param initializer: Initializer
        :param gas: Gas
        :param gas_price: Gas Price
        :return: EthereumTxSent
        """
        proxy_factory_contract = self.get_contract()
        create_proxy_fn = proxy_factory_contract.functions.createProxy(master_copy, initializer)

        tx_parameters = {
            'from': deployer_account.address
        }
        contract_address = create_proxy_fn.call(tx_parameters)

        if gas_price is not None:
            tx_parameters['gasPrice'] = gas_price

        if gas is not None:
            tx_parameters['gas'] = gas

        tx = create_proxy_fn.buildTransaction(tx_parameters)
        # Auto estimation of gas does not work. We use a little more gas just in case
        tx['gas'] = tx['gas'] + 50000
        tx_hash = self.ethereum_client.send_unsigned_transaction(tx, private_key=deployer_account.key)
        return EthereumTxSent(tx_hash, tx, contract_address)

    def deploy_proxy_contract_with_nonce(self, deployer_account: LocalAccount, master_copy: str,
                                         initializer: bytes, salt_nonce: int,
                                         gas: Optional[int] = None, gas_price: Optional[int] = None,
                                         nonce: Optional[int] = None) -> EthereumTxSent:
        """
        Deploy proxy contract via Proxy Factory using `createProxyWithNonce` (create2)
        :param deployer_account: Ethereum account
        :param master_copy: Address the proxy will point at
        :param initializer: Data for safe creation
        :param salt_nonce: Uint256 for `create2` salt
        :param gas: Gas
        :param gas_price: Gas Price
        :param nonce: Nonce
        :return: Tuple(tx-hash, tx, deployed contract address)
        """
        proxy_factory_contract = self.get_contract()
        create_proxy_fn = proxy_factory_contract.functions.createProxyWithNonce(master_copy, initializer, salt_nonce)

        tx_parameters = {
            'from': deployer_account.address
        }
        contract_address = create_proxy_fn.call(tx_parameters)

        if gas_price is not None:
            tx_parameters['gasPrice'] = gas_price

        if gas is not None:
            tx_parameters['gas'] = gas

        if nonce is not None:
            tx_parameters['nonce'] = nonce

        tx = create_proxy_fn.buildTransaction(tx_parameters)
        # Auto estimation of gas does not work. We use a little more gas just in case
        tx['gas'] = tx['gas'] + 50000
        tx_hash = self.ethereum_client.send_unsigned_transaction(tx, private_key=deployer_account.key)
        return EthereumTxSent(tx_hash, tx, contract_address)

    def get_contract(self):
        return get_proxy_factory_contract(self.ethereum_client.w3, self.address)

    def get_proxy_runtime_code(self):
        if not self.proxy_runtime_code:
            self.proxy_runtime_code = self.get_contract().functions.proxyRuntimeCode().call()
        return self.proxy_runtime_code



balancer_pool_abi = [{'inputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'constructor'},
                     {'anonymous': False,
                      'inputs': [{'indexed': True,
                                  'internalType': 'address',
                                  'name': 'src',
                                  'type': 'address'},
                                 {'indexed': True,
                                  'internalType': 'address',
                                  'name': 'dst',
                                  'type': 'address'},
                                 {'indexed': False,
                                  'internalType': 'uint256',
                                  'name': 'amt',
                                  'type': 'uint256'}],
                      'name': 'Approval',
                      'type': 'event'},
                     {'anonymous': True,
                      'inputs': [{'indexed': True,
                                  'internalType': 'bytes4',
                                  'name': 'sig',
                                  'type': 'bytes4'},
                                 {'indexed': True,
                                  'internalType': 'address',
                                  'name': 'caller',
                                  'type': 'address'},
                                 {'indexed': False,
                                  'internalType': 'bytes',
                                  'name': 'data',
                                  'type': 'bytes'}],
                      'name': 'LOG_CALL',
                      'type': 'event'},
                     {'anonymous': False,
                      'inputs': [{'indexed': True,
                                  'internalType': 'address',
                                  'name': 'caller',
                                  'type': 'address'},
                                 {'indexed': True,
                                  'internalType': 'address',
                                  'name': 'tokenOut',
                                  'type': 'address'},
                                 {'indexed': False,
                                  'internalType': 'uint256',
                                  'name': 'tokenAmountOut',
                                  'type': 'uint256'}],
                      'name': 'LOG_EXIT',
                      'type': 'event'},
                     {'anonymous': False,
                      'inputs': [{'indexed': True,
                                  'internalType': 'address',
                                  'name': 'caller',
                                  'type': 'address'},
                                 {'indexed': True,
                                  'internalType': 'address',
                                  'name': 'tokenIn',
                                  'type': 'address'},
                                 {'indexed': False,
                                  'internalType': 'uint256',
                                  'name': 'tokenAmountIn',
                                  'type': 'uint256'}],
                      'name': 'LOG_JOIN',
                      'type': 'event'},
                     {'anonymous': False,
                      'inputs': [{'indexed': True,
                                  'internalType': 'address',
                                  'name': 'caller',
                                  'type': 'address'},
                                 {'indexed': True,
                                  'internalType': 'address',
                                  'name': 'tokenIn',
                                  'type': 'address'},
                                 {'indexed': True,
                                  'internalType': 'address',
                                  'name': 'tokenOut',
                                  'type': 'address'},
                                 {'indexed': False,
                                  'internalType': 'uint256',
                                  'name': 'tokenAmountIn',
                                  'type': 'uint256'},
                                 {'indexed': False,
                                  'internalType': 'uint256',
                                  'name': 'tokenAmountOut',
                                  'type': 'uint256'}],
                      'name': 'LOG_SWAP',
                      'type': 'event'},
                     {'anonymous': False,
                      'inputs': [{'indexed': True,
                                  'internalType': 'address',
                                  'name': 'src',
                                  'type': 'address'},
                                 {'indexed': True,
                                  'internalType': 'address',
                                  'name': 'dst',
                                  'type': 'address'},
                                 {'indexed': False,
                                  'internalType': 'uint256',
                                  'name': 'amt',
                                  'type': 'uint256'}],
                      'name': 'Transfer',
                      'type': 'event'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'BONE',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'BPOW_PRECISION',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'EXIT_FEE',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'INIT_POOL_SUPPLY',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MAX_BOUND_TOKENS',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MAX_BPOW_BASE',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MAX_FEE',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MAX_IN_RATIO',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MAX_OUT_RATIO',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MAX_TOTAL_WEIGHT',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MAX_WEIGHT',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MIN_BALANCE',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MIN_BOUND_TOKENS',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MIN_BPOW_BASE',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MIN_FEE',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'MIN_WEIGHT',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'address', 'name': 'src', 'type': 'address'},
                                 {'internalType': 'address', 'name': 'dst', 'type': 'address'}],
                      'name': 'allowance',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'dst', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'amt', 'type': 'uint256'}],
                      'name': 'approve',
                      'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'address', 'name': 'whom', 'type': 'address'}],
                      'name': 'balanceOf',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'token', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'balance', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'denorm', 'type': 'uint256'}],
                      'name': 'bind',
                      'outputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'uint256',
                                  'name': 'tokenBalanceIn',
                                  'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenWeightIn', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenBalanceOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenWeightOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenAmountOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'swapFee', 'type': 'uint256'}],
                      'name': 'calcInGivenOut',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'tokenAmountIn',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'pure',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'uint256',
                                  'name': 'tokenBalanceIn',
                                  'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenWeightIn', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenBalanceOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenWeightOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenAmountIn', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'swapFee', 'type': 'uint256'}],
                      'name': 'calcOutGivenIn',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'tokenAmountOut',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'pure',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'uint256',
                                  'name': 'tokenBalanceOut',
                                  'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenWeightOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'poolSupply', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'totalWeight', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenAmountOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'swapFee', 'type': 'uint256'}],
                      'name': 'calcPoolInGivenSingleOut',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'poolAmountIn',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'pure',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'uint256',
                                  'name': 'tokenBalanceIn',
                                  'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenWeightIn', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'poolSupply', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'totalWeight', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenAmountIn', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'swapFee', 'type': 'uint256'}],
                      'name': 'calcPoolOutGivenSingleIn',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'poolAmountOut',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'pure',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'uint256',
                                  'name': 'tokenBalanceIn',
                                  'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenWeightIn', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'poolSupply', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'totalWeight', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'poolAmountOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'swapFee', 'type': 'uint256'}],
                      'name': 'calcSingleInGivenPoolOut',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'tokenAmountIn',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'pure',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'uint256',
                                  'name': 'tokenBalanceOut',
                                  'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenWeightOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'poolSupply', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'totalWeight', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'poolAmountIn', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'swapFee', 'type': 'uint256'}],
                      'name': 'calcSingleOutGivenPoolIn',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'tokenAmountOut',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'pure',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'uint256',
                                  'name': 'tokenBalanceIn',
                                  'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenWeightIn', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenBalanceOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'tokenWeightOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'swapFee', 'type': 'uint256'}],
                      'name': 'calcSpotPrice',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'spotPrice',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'pure',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'decimals',
                      'outputs': [{'internalType': 'uint8', 'name': '', 'type': 'uint8'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'dst', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'amt', 'type': 'uint256'}],
                      'name': 'decreaseApproval',
                      'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'uint256',
                                  'name': 'poolAmountIn',
                                  'type': 'uint256'},
                                 {'internalType': 'uint256[]',
                                  'name': 'minAmountsOut',
                                  'type': 'uint256[]'}],
                      'name': 'exitPool',
                      'outputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address',
                                  'name': 'tokenOut',
                                  'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'tokenAmountOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'maxPoolAmountIn', 'type': 'uint256'}],
                      'name': 'exitswapExternAmountOut',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'poolAmountIn',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address',
                                  'name': 'tokenOut',
                                  'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'poolAmountIn', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'minAmountOut', 'type': 'uint256'}],
                      'name': 'exitswapPoolAmountIn',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'tokenAmountOut',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [],
                      'name': 'finalize',
                      'outputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'address', 'name': 'token', 'type': 'address'}],
                      'name': 'getBalance',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'getColor',
                      'outputs': [{'internalType': 'bytes32', 'name': '', 'type': 'bytes32'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'getController',
                      'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'getCurrentTokens',
                      'outputs': [{'internalType': 'address[]',
                                   'name': 'tokens',
                                   'type': 'address[]'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'address', 'name': 'token', 'type': 'address'}],
                      'name': 'getDenormalizedWeight',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'getFinalTokens',
                      'outputs': [{'internalType': 'address[]',
                                   'name': 'tokens',
                                   'type': 'address[]'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'address', 'name': 'token', 'type': 'address'}],
                      'name': 'getNormalizedWeight',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'getNumTokens',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'address', 'name': 'tokenIn', 'type': 'address'},
                                 {'internalType': 'address', 'name': 'tokenOut', 'type': 'address'}],
                      'name': 'getSpotPrice',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'spotPrice',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'address', 'name': 'tokenIn', 'type': 'address'},
                                 {'internalType': 'address', 'name': 'tokenOut', 'type': 'address'}],
                      'name': 'getSpotPriceSansFee',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'spotPrice',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'getSwapFee',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'getTotalDenormalizedWeight',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'token', 'type': 'address'}],
                      'name': 'gulp',
                      'outputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'dst', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'amt', 'type': 'uint256'}],
                      'name': 'increaseApproval',
                      'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [{'internalType': 'address', 'name': 't', 'type': 'address'}],
                      'name': 'isBound',
                      'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'isFinalized',
                      'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'isPublicSwap',
                      'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'uint256',
                                  'name': 'poolAmountOut',
                                  'type': 'uint256'},
                                 {'internalType': 'uint256[]', 'name': 'maxAmountsIn', 'type': 'uint256[]'}],
                      'name': 'joinPool',
                      'outputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'tokenIn', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'tokenAmountIn', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'minPoolAmountOut', 'type': 'uint256'}],
                      'name': 'joinswapExternAmountIn',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'poolAmountOut',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'tokenIn', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'poolAmountOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'maxAmountIn', 'type': 'uint256'}],
                      'name': 'joinswapPoolAmountOut',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'tokenAmountIn',
                                   'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'name',
                      'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'token', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'balance', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'denorm', 'type': 'uint256'}],
                      'name': 'rebind',
                      'outputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address',
                                  'name': 'manager',
                                  'type': 'address'}],
                      'name': 'setController',
                      'outputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'bool', 'name': 'public_', 'type': 'bool'}],
                      'name': 'setPublicSwap',
                      'outputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'uint256',
                                  'name': 'swapFee',
                                  'type': 'uint256'}],
                      'name': 'setSwapFee',
                      'outputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'tokenIn', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'tokenAmountIn', 'type': 'uint256'},
                                 {'internalType': 'address', 'name': 'tokenOut', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'minAmountOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'maxPrice', 'type': 'uint256'}],
                      'name': 'swapExactAmountIn',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'tokenAmountOut',
                                   'type': 'uint256'},
                                  {'internalType': 'uint256', 'name': 'spotPriceAfter', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'tokenIn', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'maxAmountIn', 'type': 'uint256'},
                                 {'internalType': 'address', 'name': 'tokenOut', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'tokenAmountOut', 'type': 'uint256'},
                                 {'internalType': 'uint256', 'name': 'maxPrice', 'type': 'uint256'}],
                      'name': 'swapExactAmountOut',
                      'outputs': [{'internalType': 'uint256',
                                   'name': 'tokenAmountIn',
                                   'type': 'uint256'},
                                  {'internalType': 'uint256', 'name': 'spotPriceAfter', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'symbol',
                      'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': True,
                      'inputs': [],
                      'name': 'totalSupply',
                      'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                      'payable': False,
                      'stateMutability': 'view',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'dst', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'amt', 'type': 'uint256'}],
                      'name': 'transfer',
                      'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'src', 'type': 'address'},
                                 {'internalType': 'address', 'name': 'dst', 'type': 'address'},
                                 {'internalType': 'uint256', 'name': 'amt', 'type': 'uint256'}],
                      'name': 'transferFrom',
                      'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'},
                     {'constant': False,
                      'inputs': [{'internalType': 'address', 'name': 'token', 'type': 'address'}],
                      'name': 'unbind',
                      'outputs': [],
                      'payable': False,
                      'stateMutability': 'nonpayable',
                      'type': 'function'}]



from unittest import mock
from unittest.mock import MagicMock

from django.test import TestCase

from eth_account import Account

from .. import EthereumClient
from ..oracles import (BalancerOracle, CannotGetPriceFromOracle, CurveOracle,
                       KyberOracle, MooniswapOracle, SushiswapOracle,
                       UniswapOracle, UniswapV2Oracle, YearnOracle)
from ..oracles.oracles import (AaveOracle, PoolTogetherOracle,
                               ZerionComposedOracle)
from .ethereum_test_case import EthereumTestCaseMixin
from .utils import just_test_if_mainnet_node

gno_token_mainnet_address = '0x6810e776880C02933D47DB1b9fc05908e5386b96'
weth_token_mainnet_address = '0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2'

dai_token_mainnet_address = '0x6B175474E89094C44Da98b954EedeAC495271d0F'
usdt_token_mainnet_address = '0xdAC17F958D2ee523a2206206994597C13D831ec7'


class TestOracles(EthereumTestCaseMixin, TestCase):
    def test_kyber_oracle(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        kyber_oracle = KyberOracle(ethereum_client)
        price = kyber_oracle.get_price(gno_token_mainnet_address, weth_token_mainnet_address)
        self.assertLess(price, 1)
        self.assertGreater(price, 0)

        # Test with 2 stablecoins
        price = kyber_oracle.get_price(dai_token_mainnet_address, usdt_token_mainnet_address)
        self.assertAlmostEqual(price, 1., delta=0.5)

        price = kyber_oracle.get_price(usdt_token_mainnet_address, dai_token_mainnet_address)
        self.assertAlmostEqual(price, 1., delta=0.5)

        self.assertEqual(kyber_oracle.get_price('0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE'), 1.)

    def test_kyber_oracle_not_deployed(self):
        kyber_oracle = KyberOracle(self.ethereum_client, Account.create().address)
        random_token_address = Account.create().address
        with self.assertRaisesMessage(CannotGetPriceFromOracle, 'Cannot get price from kyber-network-proxy'):
            kyber_oracle.get_price(random_token_address)

    def test_uniswap_oracle(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        uniswap_oracle = UniswapOracle(ethereum_client)
        token_address = dai_token_mainnet_address
        price = uniswap_oracle.get_price(token_address)
        self.assertEqual(uniswap_oracle.get_uniswap_exchange.cache_info().hits, 0)
        self.assertLess(price, 1)
        self.assertGreater(price, 0)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, 'Not enough liquidity'):
            token_address = gno_token_mainnet_address
            uniswap_oracle.get_price(token_address)

        # Check batching is working
        uniswap_exchange = uniswap_oracle.get_uniswap_exchange(token_address)
        self.assertEqual(uniswap_oracle.get_uniswap_exchange.cache_info().hits, 1)
        self.assertEqual(uniswap_oracle._get_balances_using_batching(uniswap_exchange, token_address),
                         uniswap_oracle._get_balances_without_batching(uniswap_exchange, token_address))

    def test_uniswap_oracle_not_deployed(self):
        uniswap_oracle = UniswapOracle(self.ethereum_client, Account.create().address)
        random_token_address = Account.create().address
        with self.assertRaisesMessage(CannotGetPriceFromOracle, 'Non existing uniswap exchange'):
            uniswap_oracle.get_price(random_token_address)


class TestUniswapV2Oracle(EthereumTestCaseMixin, TestCase):
    @mock.patch.object(UniswapV2Oracle, 'factory_address', return_value='0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f',
                       new_callable=mock.PropertyMock)
    def test_calculate_pair_address(self, factory_address_mock: MagicMock):
        uniswap_v2_oracle = UniswapV2Oracle(self.ethereum_client)
        expected_address = '0x3e8468f66d30Fc99F745481d4B383f89861702C6'

        self.assertEqual(uniswap_v2_oracle.calculate_pair_address(gno_token_mainnet_address,
                                                                  weth_token_mainnet_address), expected_address)
        self.assertEqual(uniswap_v2_oracle.calculate_pair_address(weth_token_mainnet_address,
                                                                  gno_token_mainnet_address), expected_address)

    def test_get_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        uniswap_v2_oracle = UniswapV2Oracle(ethereum_client)

        price = uniswap_v2_oracle.get_price(gno_token_mainnet_address, weth_token_mainnet_address)
        self.assertLess(price, 1)
        self.assertGreater(price, 0)

        # Test with 2 stablecoins
        price = uniswap_v2_oracle.get_price(dai_token_mainnet_address, usdt_token_mainnet_address)
        self.assertAlmostEqual(price, 1., delta=0.5)
        self.assertEqual(uniswap_v2_oracle._decimals_cache[dai_token_mainnet_address], 18)
        self.assertEqual(uniswap_v2_oracle._decimals_cache[usdt_token_mainnet_address], 6)

        price = uniswap_v2_oracle.get_price(usdt_token_mainnet_address, dai_token_mainnet_address)
        self.assertAlmostEqual(price, 1., delta=0.5)

    def test_get_price_contract_not_deployed(self):
        uniswap_v2_oracle = UniswapV2Oracle(self.ethereum_client)
        random_token_address = Account.create().address
        with self.assertRaisesMessage(CannotGetPriceFromOracle,
                                      f'Cannot get uniswap v2 price for pair token_1={random_token_address}'):
            uniswap_v2_oracle.get_price(random_token_address)

    @mock.patch.object(UniswapV2Oracle, 'factory_address', return_value='0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f',
                       new_callable=mock.PropertyMock)
    @mock.patch.object(UniswapV2Oracle, 'get_decimals', return_value=(18, 3),
                       autospec=True)
    @mock.patch.object(UniswapV2Oracle, 'get_reserves', return_value=(int(1e20), 600),
                       autospec=True)
    def test_get_price_liquidity(self, get_reserves_mock: MagicMock, get_decimals_mock: MagicMock,
                                 factory_address_mock: MagicMock):
        uniswap_v2_oracle = UniswapV2Oracle(self.ethereum_client)
        token_1, token_2 = '0xA14F6F8867DB84a45BCD148bfaf4e4f54B4B9b12', '0xC426A8F4C79EF274Ed93faC9e1A09bFC5659B06B'

        with self.assertRaisesMessage(CannotGetPriceFromOracle, 'Not enough liquidity'):
            uniswap_v2_oracle.get_price(token_1, token_2)

        get_reserves_mock.return_value = (int(1e20), 6000)
        self.assertEqual(uniswap_v2_oracle.get_price(token_1, token_2), 0.06)

        get_reserves_mock.return_value = reversed(get_reserves_mock.return_value)
        self.assertEqual(uniswap_v2_oracle.get_price(token_2, token_1), 0.06)  # Reserves were inverted

    def test_get_pool_token_price(self):
        dai_eth_pool_address = '0xA478c2975Ab1Ea89e8196811F51A7B7Ade33eB11'
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        uniswap_v2_oracle = UniswapV2Oracle(ethereum_client)

        price = uniswap_v2_oracle.get_pool_token_price(dai_eth_pool_address)
        self.assertGreater(price, 0.)


class TestSushiSwapOracle(EthereumTestCaseMixin, TestCase):
    def test_get_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        sushiswap_oracle = SushiswapOracle(ethereum_client)

        price = sushiswap_oracle.get_price(gno_token_mainnet_address, weth_token_mainnet_address)
        self.assertLess(price, 1)
        self.assertGreater(price, 0)

        # Test with 2 stablecoins
        price = sushiswap_oracle.get_price(dai_token_mainnet_address, usdt_token_mainnet_address)
        self.assertAlmostEqual(price, 1., delta=0.5)
        self.assertEqual(sushiswap_oracle._decimals_cache[dai_token_mainnet_address], 18)
        self.assertEqual(sushiswap_oracle._decimals_cache[usdt_token_mainnet_address], 6)

        price = sushiswap_oracle.get_price(usdt_token_mainnet_address, dai_token_mainnet_address)
        self.assertAlmostEqual(price, 1., delta=0.5)


class TestAaveOracle(EthereumTestCaseMixin, TestCase):
    def test_get_token_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        uniswap_oracle = UniswapV2Oracle(ethereum_client)
        aave_oracle = AaveOracle(ethereum_client, uniswap_oracle)

        aweth_address = '0x030bA81f1c18d280636F32af80b9AAd02Cf0854e'
        price = aave_oracle.get_price(aweth_address)
        self.assertGreater(price, 0.)

        stacked_aave_address = '0x4da27a545c0c5B758a6BA100e3a049001de870f5'
        price = aave_oracle.get_price(stacked_aave_address)
        self.assertGreater(price, 0.)

        error_message = 'It is not an Aaave atoken'
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            aave_oracle.get_price(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            aave_oracle.get_price(Account.create().address)


class TestCurveOracle(EthereumTestCaseMixin, TestCase):
    def test_get_underlying_tokens(self):
        curve_token_address = '0xC25a3A3b969415c80451098fa907EC722572917F'  # Curve.fi DAI/USDC/USDT/sUSD
        local_curve_oracle = CurveOracle(self.ethereum_client)
        error_message = 'Cannot find Zerion adapter'
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            local_curve_oracle.get_underlying_tokens(curve_token_address)

        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        curve_oracle = CurveOracle(ethereum_client)

        # Curve.fi ETH/stETH (steCRV) is working with the updated adapter
        self.assertEqual(len(curve_oracle.get_underlying_tokens('0x06325440D014e39736583c165C2963BA99fAf14E')), 2)

        underlying_tokens = curve_oracle.get_underlying_tokens(curve_token_address)
        self.assertEqual(len(underlying_tokens), 4)

        self.assertCountEqual(
            [underlying_token.address for underlying_token in underlying_tokens],
            ['0x6B175474E89094C44Da98b954EedeAC495271d0F', '0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48',
             '0xdAC17F958D2ee523a2206206994597C13D831ec7', '0x57Ab1ec28D129707052df4dF418D58a2D46d5f51']
        )
        for underlying_token in underlying_tokens:
            self.assertTrue(0. < underlying_token.quantity < 1.)

        error_message = 'It is not a Zerion supported pool token'
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            curve_oracle.get_underlying_tokens(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            curve_oracle.get_underlying_tokens(Account.create().address)


class TestZerionComposedOracle(EthereumTestCaseMixin, TestCase):
    def test_zerion_composed_oracle(self):
        with self.assertRaisesMessage(ValueError, 'Expected a Zerion adapter address'):
            ZerionComposedOracle(self.ethereum_client)


class TestPoolTogetherOracle(EthereumTestCaseMixin, TestCase):
    def test_get_underlying_token(self):
        pool_together_token_address = '0xD81b1A8B1AD00Baa2D6609E0BAE28A38713872f7'  # v3 USDC Ticket
        local_pool_together_oracle = CurveOracle(self.ethereum_client)
        error_message = 'Cannot find Zerion adapter'
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            local_pool_together_oracle.get_underlying_tokens(pool_together_token_address)

        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        pool_together_oracle = PoolTogetherOracle(ethereum_client)

        underlying_tokens = pool_together_oracle.get_underlying_tokens(pool_together_token_address)
        self.assertEqual(len(underlying_tokens), 1)
        self.assertEqual(
            [underlying_token.address for underlying_token in underlying_tokens],
            ['0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48']
        )
        self.assertEqual(underlying_tokens[0].quantity, 1.)

        error_message = 'It is not a Zerion supported pool token'
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            pool_together_oracle.get_underlying_tokens(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            pool_together_oracle.get_underlying_tokens(Account.create().address)


class TestYearnOracle(EthereumTestCaseMixin, TestCase):
    def test_get_underlying_tokens(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        yearn_oracle = YearnOracle(ethereum_client)
        yearn_token_address = '0x5533ed0a3b83F70c3c4a1f69Ef5546D3D4713E44'  # Yearn Curve.fi DAI/USDC/USDT/sUSD
        yearn_underlying_token_address = '0xC25a3A3b969415c80451098fa907EC722572917F'  # Curve.fi DAI/USDC/USDT/sUSD
        iearn_token_address = '0x16de59092dAE5CcF4A1E6439D611fd0653f0Bd01'  # iearn DAI
        iearn_underlying_token_address = '0x6B175474E89094C44Da98b954EedeAC495271d0F'  # DAI
        yvault_token_address = '0xdCD90C7f6324cfa40d7169ef80b12031770B4325'  # steCRV yVault
        yvault_underlying_token_address = '0x06325440D014e39736583c165C2963BA99fAf14E'  # steCRV

        underlying_tokens = yearn_oracle.get_underlying_tokens(yearn_token_address)
        self.assertEqual(len(underlying_tokens), 1)
        underlying_token = underlying_tokens[0]
        self.assertEqual(underlying_token.quantity, 1.)
        self.assertEqual(underlying_token.address, yearn_underlying_token_address)

        underlying_tokens = yearn_oracle.get_underlying_tokens(iearn_token_address)
        self.assertEqual(len(underlying_tokens), 1)
        underlying_token = underlying_tokens[0]
        self.assertEqual(underlying_token.quantity, 1.)
        self.assertEqual(underlying_token.address, iearn_underlying_token_address)

        underlying_tokens = yearn_oracle.get_underlying_tokens(yvault_token_address)
        self.assertEqual(len(underlying_tokens), 1)
        underlying_token = underlying_tokens[0]
        self.assertEqual(underlying_token.quantity, 1.)
        self.assertEqual(underlying_token.address, yvault_underlying_token_address)

        # Test yToken
        y_token = '0x30FCf7c6cDfC46eC237783D94Fc78553E79d4E9C'
        yearn_underlying_token = '0x3a664Ab939FD8482048609f652f9a0B0677337B9'
        underlying_tokens = yearn_oracle.get_underlying_tokens(y_token)
        self.assertEqual(len(underlying_tokens), 1)
        underlying_token = underlying_tokens[0]
        self.assertAlmostEqual(underlying_token.quantity, 1., delta=0.5)
        self.assertEqual(underlying_token.address, yearn_underlying_token)

        error_message = 'It is not a Yearn yToken/yVault'
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            yearn_oracle.get_underlying_tokens(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            yearn_oracle.get_underlying_tokens(Account.create().address)


class TestBalancerOracle(EthereumTestCaseMixin, TestCase):
    def test_get_pool_token_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        uniswap_oracle = UniswapV2Oracle(ethereum_client)
        balancer_oracle = BalancerOracle(ethereum_client, uniswap_oracle)
        balancer_token_address = '0x59A19D8c652FA0284f44113D0ff9aBa70bd46fB4'  # Balancer 80% BAL + 20% WETH

        price = balancer_oracle.get_pool_token_price(balancer_token_address)
        self.assertAlmostEqual(price, 1., delta=0.9)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, 'It is not a balancer pool token'):
            balancer_oracle.get_pool_token_price(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, 'It is not a balancer pool token'):
            balancer_oracle.get_pool_token_price(Account.create().address)


class TestMooniswapOracle(EthereumTestCaseMixin, TestCase):
    def test_get_pool_token_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        uniswap_oracle = UniswapV2Oracle(ethereum_client)
        mooniswap_oracle = MooniswapOracle(ethereum_client, uniswap_oracle)
        mooniswap_pool_address = '0x6a11F3E5a01D129e566d783A7b6E8862bFD66CcA'  # 1inch Liquidity Pool (ETH-WBTC)
        other_version_mooniswap_pool_address = '0x8a2f2F8637deb4Ee7C9400A240d27E3A32147dA4'  # Mooniswap V1 (OWL-GNO)

        price = mooniswap_oracle.get_pool_token_price(mooniswap_pool_address)
        self.assertGreater(price, 0)

        price = mooniswap_oracle.get_pool_token_price(other_version_mooniswap_pool_address)
        self.assertGreater(price, 0)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, 'It is not a mooniswap pool token'):
            mooniswap_oracle.get_pool_token_price(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, 'It is not a mooniswap pool token'):
            mooniswap_oracle.get_pool_token_price(Account.create().address)



# flake8: noqa F401
from .ethereum_client import (EthereumClient, EthereumClientProvider,
                              EthereumNetwork, FromAddressNotFound,
                              GasLimitExceeded, InsufficientFunds,
                              InvalidERC20Info, InvalidERC721Info,
                              InvalidNonce, NonceTooHigh, NonceTooLow,
                              ParityTraceDecodeException,
                              ReplacementTransactionUnderpriced,
                              SenderAccountNotFoundInNode,
                              TransactionAlreadyImported,
                              TransactionQueueLimitReached, UnknownAccount)



from typing import List, Tuple, Union

from ethereum.utils import checksum_encode, ecrecover_to_pub, sha3
from hexbytes import HexBytes


def signature_split(signatures: Union[bytes, str], pos: int = 0) -> Tuple[int, int, int]:
    """
    :param signatures: signatures in form of {bytes32 r}{bytes32 s}{uint8 v}
    :param pos: position of the signature
    :return: Tuple with v, r, s
    """
    signatures = HexBytes(signatures)
    signature_pos = 65 * pos
    if len(signatures[signature_pos:signature_pos + 65]) < 65:
        raise ValueError(f'Signature must be at least 65 bytes {signatures.hex()}')
    r = int.from_bytes(signatures[signature_pos:32 + signature_pos], 'big')
    s = int.from_bytes(signatures[32 + signature_pos:64 + signature_pos], 'big')
    v = signatures[64 + signature_pos]

    return v, r, s


def signature_to_bytes(v: int, r: int, s: int) -> bytes:
    """
    Convert ecdsa signature to bytes
    :param v:
    :param r:
    :param s:
    :return: signature in form of {bytes32 r}{bytes32 s}{uint8 v}
    """

    byte_order = 'big'

    return (r.to_bytes(32, byteorder=byte_order)
            + s.to_bytes(32, byteorder=byte_order)
            + v.to_bytes(1, byteorder=byte_order))


def signatures_to_bytes(signatures: List[Tuple[int, int, int]]) -> bytes:
    """
    Convert signatures to bytes
    :param signatures: list of tuples(v, r, s)
    :return: 65 bytes per signature
    """
    return b''.join([signature_to_bytes(v, r, s) for v, r, s in signatures])


def get_signing_address(signed_hash: Union[bytes, str], v: int, r: int, s: int) -> str:
    """
    :return: checksummed ethereum address, for example `0x568c93675A8dEb121700A6FAdDdfE7DFAb66Ae4A`
    :rtype: str
    """
    encoded_64_address = ecrecover_to_pub(HexBytes(signed_hash), v, r, s)
    address_bytes = sha3(encoded_64_address)[-20:]
    return checksum_encode(address_bytes)
